from pyPhases import Phase
from glob import glob
import json
from pathlib import Path
import pickle
import pandas as pd
from pyPhases import pdict
import fnmatch


class LogOverview(Phase):
    def loadConfig(self, json_path):
        with open(json_path, "r") as file:
            config = json.load(file)
        return config

    def flatten_with_index(self, lis, names):
        for i, sublist in enumerate(lis):
            if isinstance(sublist, list):
                for item in sublist:
                    yield f"{item}_{names[i]}"
            else:
                yield sublist

    def findBestResults(self, logFolderPath, config):
        # load from log.csv if its exists or from checkpoint-files inside the folder
        logFile = "log.csv"
        csvPath = Path(logFolderPath) / logFile
        metrics = config["trainingParameter"]["validationMetrics"]
        labelNames = config["classification"]["labelNames"]
        labelNames += self.getConfig("overview.addLabelnames", [])
        
        # Add last modified date of log file
        results = {}
        if csvPath.exists():
            # Get last modified time of log file
            last_modified = csvPath.stat().st_mtime
            results["last_modified"] = pd.to_datetime(last_modified, unit='s')
            
            prefix = "val_"
            # read Header and select best "auprc"-col
            df = pd.read_csv(str(csvPath))
            if self.useEpoch:
                results.update({
                    m: df[m].iloc[:self.useEpoch].max()
                    for m in df.columns if m.startswith(prefix)
                })
            else:
                results.update({
                    m: df[m].max()
                    for m in df.columns
                })
            results["epochs"] = len(df)
        else:
            flattened_metrics = list(self.flatten_with_index(metrics[0 : len(labelNames)], labelNames))

            results = {m: 0 for m in flattened_metrics}
            maxEpoch = 0
            for checkPointFile in glob(logFolderPath + "/checkpoint*"):
                split = ".".join(checkPointFile.split("/")[-1].split(".")[:-1]).split("_")
                epoch = split[1]
                epochResults = split[2:]
                results = {m: max(results[m], float(epochResults[i])) for i, m in enumerate(flattened_metrics)}
                maxEpoch = max(maxEpoch, int(epoch))
                results["epochs"] = maxEpoch
                
                # Get last modified time of the latest checkpoint file
                checkpoint_path = Path(checkPointFile)
                if checkpoint_path.exists():
                    last_modified = checkpoint_path.stat().st_mtime
                    results["last_modified"] = pd.to_datetime(last_modified, unit='s')
                
        return results


    def findSegmentResults(self, evalFolder, prefix):
        dataId = evalFolder.split("/")[-1]
        dataIdSegments = "-".join(dataId.split("-")[:-2])
        evalResults = list(glob(f"data/evalResults{dataIdSegments}*"))
        results = {}
        if len(evalResults) > 1:
            raise Exception("there should only be one results")
        if len(evalResults) == 1:
            with open(evalResults[0], "rb") as f:
                data = pickle.load(f)
                results = {
                    prefix + "auprc_seg": data[1]["auprc"],
                    prefix + "f1_seg": data[1]["f1"],
                }
        return results

    def findEvalResults(self, evalFolder):
        resultFile = "recordResultsEvents.csv"
        csvPath = evalFolder + "/" + resultFile
        prefix = "eval"
        results = {}

        if Path(csvPath).exists():
            df = pd.read_csv(str(csvPath))
            values = list(df.columns)
            exampleRecordId = df.iloc[1, 0]
            if exampleRecordId.startswith("acq_"):
                prefix = "tsm_"
            elif exampleRecordId.startswith("shhs1"):
                prefix = "shhs1"
            elif exampleRecordId.startswith("mros"):
                prefix = "mros"
            elif exampleRecordId.startswith("mesa"):
                prefix = "mesa"
            if "examples" in values:
                values.remove("examples")

            df["f1"].mean()

            meanValues = ["f1"]
            suffix = "_mean"
            results = {f"{prefix}_{m}_{suffix}": df[m].mean() for m in meanValues}
            hasIndex = "indexArousal-prediction" in values
            medValues = ["eventCountDiff"]
            if "countArousal-software" in values:
                df["arCount-diff"] = df["countArousal-prediction"] - df["countArousal-software"]

                if hasIndex:
                    df["arI-diff"] = df["indexArousal-prediction"] - df["indexArousal-software"]
            elif "countArousal-prediction" in values:
                df["arCount-diff"] = df["countArousal-prediction"] - df["countArousal-truth"]
                if hasIndex:
                    df["arI-diff"] = df["indexArousal-prediction"] - df["indexArousal-truth"]
                medValues.append("arCount-diff")

            if hasIndex:
                medValues.append("arI-diff")

            suffix = "_med"
            results |= {f"{prefix}_{m}_{suffix}": df[m].median() for m in medValues}

        segmentResults = self.findSegmentResults(evalFolder, prefix)
        results |= segmentResults

        return results

    def find_different_keys(self, dict1, dict2, parent_key="", different_keys=None):
        if different_keys is None:
            different_keys = []

        for key in dict1.keys():
            value1 = dict1[key]
            value2 = dict2.get(key)

            if isinstance(value1, dict) and value2 and isinstance(value2, dict):
                self.find_different_keys(value1, value2, parent_key + "." + key, different_keys)
            elif value1 != value2:
                different_keys.append(parent_key + "." + key)

        return different_keys

    def getNestedKeys(self, myPdict, path=""):
        keys = []

        for k, value in myPdict.items():
            if isinstance(value, dict):
                keys += self.getNestedKeys(value, path=path + "." + k)
            else:
                keys.append(path + "." + k)
        if path == "":
            keys = [k[1:] for k in keys]
        return keys

    def shouldIgnoreValue(self, key, value, ignoreValues):
        """Check if a key-value pair should be ignored based on ignoreValues config"""
        for ignorePattern in ignoreValues:
            if isinstance(ignorePattern, dict):
                for ignoreKey, ignoreVal in ignorePattern.items():
                    # Support wildcard matching for keys
                    if fnmatch.fnmatch(key, ignoreKey):
                        # Convert value to string for comparison if it's a list
                        compareValue = str(value) if isinstance(value, list) else value
                        # Support wildcard matching for values
                        if isinstance(ignoreVal, str) and fnmatch.fnmatch(str(compareValue), ignoreVal):
                            return True
                        elif compareValue == ignoreVal:
                            return True
            elif isinstance(ignorePattern, str):
                # Support legacy format: "key:value"
                if ":" in ignorePattern:
                    ignoreKey, ignoreVal = ignorePattern.split(":", 1)
                    if fnmatch.fnmatch(key, ignoreKey.strip()):
                        compareValue = str(value) if isinstance(value, list) else value
                        if fnmatch.fnmatch(str(compareValue), ignoreVal.strip()):
                            return True
        return False

    def calculateDistinctKeys(self, logResults, ignoreKeys, ignoreValues):
        """Calculate distinct key values from the filtered log results"""
        distinctKeyValues = {}
      
        # First pass: collect all possible keys
        allKeys = set()
        for config, results in logResults:
            keys = self.getNestedKeys(config)
            for key in keys:
                if not any(fnmatch.fnmatch(key, pattern) for pattern in ignoreKeys):
                    allKeys.add(key)
      
        # Second pass: collect values for all keys, including None for missing keys
        for config, results in logResults:
            for key in allKeys:
                try:
                    value = config[key.split(".")]
                    if isinstance(value, dict):
                        continue
                    # Skip key-value pairs that should be ignored
                    if self.shouldIgnoreValue(key, value, ignoreValues):
                        continue
                    if isinstance(value, list):
                        value = str(value)
                except:
                    value = None
              
                if key not in distinctKeyValues:
                    distinctKeyValues[key] = set()
                distinctKeyValues[key].add(value)
      
        return distinctKeyValues

    def convert_to_categories(self, df):
        """Convert all columns to appropriate types for correlation analysis"""
        df_converted = df.copy()
        
        for col in df_converted.columns:
            # Skip if column is already numeric
            if pd.api.types.is_numeric_dtype(df_converted[col]):
                continue
                
            # Handle None/NaN values
            if df_converted[col].isna().all():
                # If all values are NaN, convert to numeric with NaN
                df_converted[col] = pd.to_numeric(df_converted[col], errors='coerce')
                continue
            
            # Get unique non-null values
            # unique_values = df_converted[col].to(str).dropna().unique()
            
            # Try to convert to numeric first
            numeric_series = pd.to_numeric(df_converted[col], errors='coerce')
            
            # If conversion was successful (not all NaN after conversion)
            if not numeric_series.isna().all() and not (numeric_series.isna() & df_converted[col].notna()).any():
                df_converted[col] = numeric_series
            else:
                # Convert to categorical with numeric codes
                # First, convert to string to handle mixed types
                df_converted[col] = df_converted[col].astype(str)
                # Replace 'nan' string back to actual NaN
                df_converted[col] = df_converted[col].replace('nan', pd.NA)
                # Convert to category
                df_converted[col] = df_converted[col].astype('category')
                # Get category codes (this creates numeric representation)
                df_converted[col] = df_converted[col].cat.codes
                # Replace -1 (which represents NaN in category codes) with actual NaN
                df_converted[col] = df_converted[col].replace(-1, pd.NA)
                # Convert to float to handle NaN properly
                df_converted[col] = df_converted[col].astype('float64')
        
        return df_converted

    def apply_value_replacements(self, df, replace_config):
        """Apply value replacements based on the replaceValue configuration"""
        if not replace_config:
            return df
        
        df_replaced = df.copy()
        
        for column, replacements in replace_config.items():
            if column in df_replaced.columns:
                # Handle regular value replacements
                df_replaced[column] = df_replaced[column].replace(replacements)
                
                # Handle "default" value for None/null values
                if "default" in replacements:
                    default_value = replacements["default"]
                    df_replaced[column] = df_replaced[column].fillna(default_value)
            else:
                print(f"Warning: Column '{column}' not found in dataframe for value replacement")
        
        return df_replaced


    def main(self):
        logDirectory = self.getConfig("overview.logDirectory", "results/")
        evalFolder = self.getConfig("overview.evalFolder", "eval/")
        configName = self.getConfig("overview.configName", "project.config")
        minEpochs = self.getConfig("overview.minEpochs", 10)
        ignoreKeys = self.getConfig("overview.ignore", [])
        ignoreValues = self.getConfig("overview.ignoreValues", [])
        queryFilter = self.getConfig("overview.query", False)
        dropColumns = self.getConfig("overview.dropColumns", [])
        replaceValues = self.getConfig("overview.replaceValues", {})
        self.useEpoch = self.getConfig("overview.useEpoch", -1)
        minEpochs = max(self.useEpoch, minEpochs)

        logResults = []
        # iterate over all log folders
        for folder in glob(logDirectory + "/*"):
            # get Config
            modelConfig = folder + "/" + configName
            if not Path(modelConfig).exists():
                subFiles = list(glob(folder + "/*"))
                if len(subFiles) > 0:
                    print("no config found for %s: " % folder)
                    print("files: %s" % (subFiles))
                continue
            config = {"log-path": folder}
            config |= self.loadConfig(modelConfig)

            # read best results
            bestResults = self.findBestResults(folder, config)
            modelStateId = folder[len(logDirectory) :]
            config["modelStateId"] = modelStateId
            evalFolders = list(glob(f"{evalFolder}*{modelStateId}*"))
            config["evalFolders"] = evalFolders
            for f in evalFolders:
                bestResults |= self.findEvalResults(f)

            logResults.append((pdict(config), bestResults))

        def resultFilter(configResult):
            return (
                "epochs" in configResult[1] and configResult[1]["epochs"] is not None and configResult[1]["epochs"] > minEpochs
            )

        logResults = list(filter(resultFilter, logResults))

        # Apply query filter if specified
        if queryFilter:
            # First, create a temporary dataframe with all keys to apply the query
            tempDistinctKeyValues = self.calculateDistinctKeys(logResults, [], [])  # No ignores for temp calculation
            tempRows = []
            for config, results in logResults:
                tempRelevantConfig = {
                    k: config[k.split(".")] if k.split(".") in config else None for k in tempDistinctKeyValues.keys()
                }
                tempRow = tempRelevantConfig | results
                tempRows.append(tempRow)

            tempDf = pd.DataFrame(tempRows)

            # Apply the query filter
            filteredDf = tempDf.query(queryFilter.replace("\n", ""))

            # Get the indices of rows that passed the filter
            filteredIndices = filteredDf.index.tolist()

            # Filter logResults based on the query results
            logResults = [logResults[i] for i in filteredIndices]

            print(f"Query '{queryFilter}' filtered {len(tempDf)} rows down to {len(logResults)} rows")


        # Recalculate distinct keys after filtering
        distinctKeyValues = self.calculateDistinctKeys(logResults, ignoreKeys, ignoreValues)

        # Build final dataframe
        rows = []
        for config, results in logResults:
            relevantConfig = {k: config[k.split(".")] if k.split(".") in config else None for k, values in distinctKeyValues.items() if len(values) > 1}
            row = relevantConfig | results
            rows.append(row)
        df = pd.DataFrame(rows)

        # Apply value replacements
        df = self.apply_value_replacements(df, replaceValues)

        df = df.drop(columns=dropColumns, errors="ignore")
        
        
        # Convert all columns to numeric/categorical for correlation analysis
        df_converted = self.convert_to_categories(df)
        
        # Save both original and converted dataframes
        df.to_csv(logDirectory + "/logOverview.csv", index=False)
        df_converted.to_csv(logDirectory + "/logOverview_categorical.csv", index=False)
        
        # Now you can compute correlations on df_converted
        correlation_matrix = df_converted.corr()
        correlation_matrix.to_csv(logDirectory + "/correlation_matrix.csv")
        
        print(f"Saved original data to: {logDirectory}/logOverview.csv")
        print(f"Saved categorical data to: {logDirectory}/logOverview_categorical.csv")
        print(f"Saved correlation matrix to: {logDirectory}/correlation_matrix.csv")




# Setup project for unittest discovery
from phases.commands.test import Test as PhasesTest
from pyPhases.test import TestCase


def _setup_project():
    """Setup the project for testing - called when tests module is imported."""
    if TestCase.project is None:
        phasesTest = PhasesTest({})
        phasesTest.prepareConfig()
        testConfig = phasesTest.loadConfig("tests/config.yml")
        phasesTest.config.update(testConfig)
        project = phasesTest.createProjectFromConfig(phasesTest.config)
        TestCase.project = project


_setup_project()



import json
from pathlib import Path
from typing import Optional, Dict, List

import numpy as np
from pyPhases import Phase
from pyPhasesML import Model, ModelManager
from pyPhasesRecordloader import RecordLoader
from pyPhasesRecordloader.RecordLoader import ChannelsNotPresent

from SleePyPhases.DataManipulation import DataManipulation
from SleePyPhases.FeatureExtraction import FeatureExtraction
from SleePyPhases.PreManipulation import PreManipulation
from SleePyPhases.SignalPreprocessing import SignalPreprocessing


class Predict(Phase):
    """
    Execute the trained model on a given EDF file for prediction.
    
    Configuration options:
    - predict.inputFile: Path to the input EDF file (required)
    - predict.weights: Path to custom model weights (optional, uses trained weights by default)
    - predict.recordLoader: Name of the RecordLoader class to use (default: EDFRecordLoader)
    - predict.channelMapping: Dict mapping expected channels to actual channels in the file
    - predict.lastChannelMappingPath: Path to save/load last channel mapping (default: lastchannelmapping.json)
    """
    
    CHANNEL_MAPPING_FILE = "lastchannelmapping.json"
    
    def getModel(self, weightsPath: Optional[str] = None) -> Model:
        """Load the model with optional custom weights."""
        if weightsPath:
            # Load custom weights from specified path
            model = ModelManager.getModel(True)
            model.build()
            model.load(weightsPath)
        else:
            # Use default trained model state
            modelState = self.project.getData("modelState", Model, generate=False)
            model = ModelManager.getModel(True)
            model.build()
            model.loadState(modelState)
        return model
    
    def loadChannelMapping(self) -> Dict[str, str]:
        """Load the last used channel mapping from file."""
        mappingPath = self.getConfig("predict.lastChannelMappingPath", self.CHANNEL_MAPPING_FILE)
        if Path(mappingPath).exists():
            with open(mappingPath, 'r') as f:
                return json.load(f)
        return {}
    
    def saveChannelMapping(self, mapping: Dict[str, str]):
        """Save the channel mapping to file."""
        mappingPath = self.getConfig("predict.lastChannelMappingPath", self.CHANNEL_MAPPING_FILE)
        with open(mappingPath, 'w') as f:
            json.dump(mapping, f, indent=2)
    
    def promptChannelMapping(self, missingChannels: List[str], availableChannels: List[str], 
                              defaultMapping: Dict[str, str]) -> Dict[str, str]:
        """Interactively prompt user to map missing channels."""
        print("\n=== Channel Mapping Required ===")
        print(f"Available channels in file: {availableChannels}")
        print(f"Missing expected channels: {missingChannels}")
        print()
        
        mapping = {}
        for channel in missingChannels:
            default = defaultMapping.get(channel, "")
            if default:
                prompt = f"Map '{channel}' to (default: {default}): "
            else:
                prompt = f"Map '{channel}' to: "
            
            userInput = input(prompt).strip()
            if userInput == "" and default:
                mapping[channel] = default
            elif userInput:
                mapping[channel] = userInput
            else:
                print(f"Warning: No mapping provided for '{channel}', skipping...")
        
        return mapping
    
    def getSignalFromFile(self, inputFile: str, channelMapping: Optional[Dict[str, str]] = None):
        """Load and preprocess the signal from the input file."""
        loaderName = self.getConfig("predict.recordLoader", False)
        if loaderName:
            self.setConfig("loader.recordLoader", loaderName)
        
        recordLoader = RecordLoader.get()
        
        # Get preprocessing configuration
        targetChannels = self.getConfig("preprocessing.targetChannels", [])
        expectedChannels = [ch[0] if isinstance(ch, list) else ch for ch in targetChannels]
        
        # Apply channel mapping if provided
        if channelMapping:
            recordLoader.chanelNameAliasMap = {v: k for k, v in channelMapping.items()}
        
        try:
            recordSignal = recordLoader.loadSignal(inputFile)
        except ChannelsNotPresent as e:
            # Handle missing channels with interactive mapping
            missingChannels = e.channels if hasattr(e, 'channels') else []
            availableChannels = recordSignal.signalNames
            
            # Load default mapping from config or last saved mapping
            defaultMapping = self.getConfig("predict.channelMapping", {})
            savedMapping = self.loadChannelMapping()
            defaultMapping.update(savedMapping)
            
            # Prompt user for mapping
            newMapping = self.promptChannelMapping(missingChannels, availableChannels, defaultMapping)
            
            if newMapping:
                # Save the mapping for future use
                fullMapping = {**defaultMapping, **newMapping}
                self.saveChannelMapping(fullMapping)
                
                # Retry with new mapping
                recordLoader.chanelNameAliasMap = {v: k for k, v in fullMapping.items()}
                recordSignal, events = recordLoader.loadRecord(inputFile)
            else:
                raise
        
        return recordSignal
    
    def preprocessSignal(self, recordSignal, targetSignals=None):
        """Preprocess the signal for model input."""
        preprocessingConfig = self.getConfig("preprocessing")
        targetFrequency = preprocessingConfig["targetFrequency"]
        targetSignals = preprocessingConfig["targetChannels"] if targetSignals is None else None
        
        featureChannels = preprocessingConfig.get("featureChannels", [])
        
        recordSignal.targetFrequency = targetFrequency
        
        # Get signal names
        signalNames = [recordSignal.getFirstSignalName(s) for s in targetSignals]
        recordSignal = recordSignal[signalNames]
        
        # Apply preprocessing
        signalProcessing = SignalPreprocessing.getInstance(preprocessingConfig)
        signalProcessing.preprocessingSignal(recordSignal)
        
        # Apply pre-manipulation
        preDataManipulation = PreManipulation.getInstance(preprocessingConfig.get("manipulationSteps", []))
        recordSignal, _ = preDataManipulation(recordSignal, [])

        # Extract features if configured
        if featureChannels:
            featureExtraction = FeatureExtraction.getInstance(self.project)
            featureExtraction.extractChannelsByConfig(recordSignal, featureChannels)

        processedSignal = recordSignal.getSignalArray(targetSignals)
        return processedSignal

    def predict(self, inputFile: str) -> np.ndarray:
        """Run prediction on the input file."""
        # Get channel mapping from config
        channelMapping = self.getConfig("predict.channelMapping", None)

        # Load and preprocess the signal
        recordSignal = self.getSignalFromFile(inputFile, channelMapping)

        return self.predictRecordSignal(recordSignal)

    def predictRecordSignal(self, recordSignal, dummyY=None):
        processedSignal = self.preprocessSignal(recordSignal)

        # # Prepare input for model
        # if len(processedSignal.shape) == 2:
        #     processedSignal = processedSignal.reshape([1] + list(processedSignal.shape))

        # Apply segment manipulation for evaluation
        manipulationSteps = self.getConfig("segmentManipulationEval", [])
        da = DataManipulation.getInstance(manipulationSteps, "test", self.project.config)
        dummyY = np.array([0]).reshape(1, -1, 1) if  dummyY is None else dummyY
        processedSignal, dummyY = da((processedSignal, dummyY))  # Dummy label

        with self.project:
            overwriteConfig = self.getConfig("predict", {})
            self.project.updateConfig(overwriteConfig)

            # Load model with optional custom weights
            weightsPath = self.getConfig("predict.weights", False)
            model = self.getModel(weightsPath)

            # Run prediction
            prediction = model.predict(processedSignal, returnNumpy=True)

            # Apply post-prediction manipulation if configured
            manipulationAfterPredict = self.getConfig("manipulationAfterPredict", False)
            if manipulationAfterPredict:
                threshold = self.getConfig("fixedThreshold", False)
                threshold = threshold or self.getData("threshold", list, generate=False)
                da = DataManipulation.getInstance(manipulationAfterPredict, "test", self.project.config, threshold=threshold)
                prediction, _ = da((prediction, dummyY))

        return prediction

    def generateData(self, name):
        if name == "prediction":
            inputFile = self.getConfig("predict.inputFile")
            if not inputFile:
                raise ValueError("predict.inputFile must be specified in config")

            prediction = self.predict(inputFile)
            self.registerData("prediction", prediction)
            return prediction

    def main(self):
        inputFile = self.getConfig("predict.inputFile")
        if not inputFile:
            self.logError("predict.inputFile must be specified in config")
            return

        if not Path(inputFile).exists():
            self.logError(f"Input file not found: {inputFile}")
            return

        self.logSuccess(f"Running prediction on: {inputFile}")

        prediction = self.predict(inputFile)

        # Get label names for output
        labelNames = self.getConfig("classification.labelNames", ["prediction"])
        classNames = self.getConfig("classification.classNames", [])

        self.logSuccess(f"Prediction complete. Output shape: {prediction.shape}")

        # Display summary of predictions
        for i, labelName in enumerate(labelNames):
            if isinstance(prediction, list):
                pred = prediction[i]
            elif len(prediction.shape) > 1 and prediction.shape[-1] > 1:
                pred = prediction[..., i] if i < prediction.shape[-1] else prediction
            else:
                pred = prediction

            if len(classNames) > i:
                classes = classNames[i]
                if len(pred.shape) > 1:
                    pred_classes = np.argmax(pred, axis=-1)
                else:
                    pred_classes = pred
                unique, counts = np.unique(pred_classes, return_counts=True)
                self.logSuccess(f"{labelName} distribution:")
                for u, c in zip(unique, counts):
                    className = classes[int(u)] if int(u) < len(classes) else f"class_{u}"
                    self.logSuccess(f"  {className}: {c} samples ({100*c/len(pred_classes):.1f}%)")

        return prediction




from unittest import TestCase
from unittest.mock import MagicMock

import numpy as np
import numpy.testing as npt
from pyPhasesRecordloader import Event, RecordSignal, Signal

from SleePyPhases.phases.Extract import RecordProcessor


class TestRecordProcessor(TestCase):
    def getRecordProcessor(self):
        recordLoader = MagicMock()
        extractor = MagicMock()

        eventManager = MagicMock()
        eventManager.getEventSignalFromList.return_value = {"arousal": np.array([0, 1, 2, 3, 4])}
        stepsPerType = {"test": ["S1", "S2"]}
        preprocessingConfig = {
            "targetFrequency": 1,
            "labelFrequency": 1,
            "dtype": "float16",
            "targetChannels": ["C1", "C2", "C3"],
            "stepsPerType": stepsPerType,
            "extendEvents": {"test": [0, 1]},
        }

        featureExtraction = MagicMock()
        preDataManipulation = MagicMock()

        return RecordProcessor(
            recordLoader,
            preprocessingConfig,
            extractor,
            eventManager,
            labelChannels=["SleepArousals"],
            featureExtraction=featureExtraction,
            preDataManipulation=preDataManipulation,
        )

    def testCutFirstAndLastWakeStartSPT(self):
        events = [
            Event("arousal", 0, 10),
            Event("W", 30, 30),
            Event("N1", 60, 30),
            Event("arousal", 80, 10),
            Event("arousal", 85, 10),
        ]
        recordProcessor = self.getRecordProcessor()
        rs = RecordSignal(1)
        rs.addSignal(Signal("test", np.arange(100), frequency=1))
        cuttedEvents = recordProcessor.tailorToSleepScoring(rs, events, useSPT=True)

        # self.assertEqual(len(cuttedEvents), 3)
        self.assertEqual(
            cuttedEvents,
            [
                Event("N1", 0, 30),
                Event("arousal", 20, 10),
                Event("arousal", 25, 5),
            ],
        )
        npt.assert_equal(rs.signals[0].signal, np.arange(60, 90))
        
    def testCutFirstAndLastWakeStart(self):
        events = [
            Event("arousal", 0, 10),
            Event("W", 30, 30),
            Event("N1", 60, 30),
            Event("arousal", 80, 10),
            Event("arousal", 85, 10),
        ]
        recordProcessor = self.getRecordProcessor()
        rs = RecordSignal(1)
        rs.addSignal(Signal("test", np.arange(100), frequency=1))
        cuttedEvents = recordProcessor.tailorToSleepScoring(rs, events, useSPT=False)

        # self.assertEqual(len(cuttedEvents), 3)
        self.assertEqual(
            cuttedEvents,
            [
                Event("W", 0, 30),
                Event("N1", 30, 30),
                Event("arousal", 50, 10),
                Event("arousal", 55, 5),
            ],
        )
        npt.assert_equal(rs.signals[0].signal, np.arange(30, 90))

    def testCutFirstAndLastWakeEndSPT(self):
        events = [
            Event("arousal", 65, 5),
            Event("N1", 60, 20),
            Event("W", 80, 30),
            Event("arousal", 80, 40),
        ]
        recordProcessor = self.getRecordProcessor()
        rs = RecordSignal(1)
        rs.addSignal(Signal("test", np.arange(100), frequency=1))

        cuttedEvents = recordProcessor.tailorToSleepScoring(rs, events, useSPT=True)

        self.assertEqual(
            cuttedEvents,
            [
                Event("arousal", 5, 5),
                Event("N1", 0, 20),
            ],
        )
        npt.assert_equal(rs.signals[0].signal, np.arange(60, 80))

    def testCutFirstAndLastWakeEnd(self):
        events = [
            Event("arousal", 65, 5),
            Event("N1", 60, 20),
            Event("W", 80, 30),
            Event("arousal", 80, 40),
        ]
        recordProcessor = self.getRecordProcessor()
        rs = RecordSignal(1)
        rs.addSignal(Signal("test", np.arange(120), frequency=1))
        
        cuttedEvents = recordProcessor.tailorToSleepScoring(rs, events, useSPT=False)

        self.assertEqual(
            cuttedEvents,
            [
                Event("arousal", 5, 5),
                Event("N1", 0, 20),
                Event("W", 20, 30),
                Event("arousal", 20, 30),
            ],
        )
        npt.assert_equal(rs.signals[0].signal, np.arange(60, 110))

    def testLightEvents(self):
        events = [
            Event("arousal", 65, 5),
            Event("lightOff", 75),
            Event("lightOn", 85),
            Event("N1", 60, 20),
            Event("W", 80, 30),
            Event("arousal", 80, 40),
        ]
        recordProcessor = self.getRecordProcessor()
        rs = RecordSignal(1)
        rs.addSignal(Signal("test", np.arange(120), frequency=1))
        
        cuttedEvents = recordProcessor.tailorToSleepScoring(rs, events, useSPT=False)

        self.assertEqual(
            cuttedEvents,
            [
                Event("N1", 0, 5),
                Event("W", 5, 5),
                Event("arousal", 5, 5),
            ],
        )
        npt.assert_equal(rs.signals[0].signal, np.arange(75, 85))



import contextlib
from unittest.mock import MagicMock
import numpy as np
import numpy.testing as npt
import pandas as pd
from pyPhases.test.Mocks import OverwriteConfig
from pyPhases.test.TestCase import TestCase
from pyPhasesRecordloader import RecordSignal, Signal, RecordLoader, SignalPreprocessing

from SleePyPhases.phases.BuildDataset import BuildDataset


class TestBuildDataset(TestCase):
    """ Test Scenario:
        - Dataset with 10 Recordings: training [id 0-3], validation [id 4-6], test [id 7-9]
        - Each Recording just consist of 2 channels [[0,1], [2,3], ... , [16,17], [18,19]]
        - DataManipulation: transform each recording to fixed size of 10 (0 padded): [[[0, 0, 0, 0, 0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 1, 0, 0, 0, 0]], ...]

    """
    phase = BuildDataset()

    C1 = np.array([1, 2, 3, 4, 5])
    C2 = np.array([6, 7, 8, 9, 10])
    C3 = C2

    def config(self):
        self.project.setConfig("dataversion.split", {})
        return {
            "useLoader": "test",
            "dataversion": {
                "split": {
                    "training": ["0:4"],
                    "validation": ["4:7"],
                    "test": ["7:10"],
                },
                "seed": None,
                "recordIds": ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"],
                "groupBy": None,
            },
            "segmentManipulation": [{"name": "addBatchDimension"}, {"name": "fixedSize", "position": "center", "size": 10}],
            "segmentManipulationEval": [{"name": "addBatchDimension"}, {"name": "fixedSize", "position": "center", "size": 10}],
            "evalOn": None,
        }

    def setUp(self):
        super().setUp()

        dataProcessed = np.arange(20).reshape(10, 1, 2)
        dataFeatures = np.arange(10).reshape(10, 1, 1)

        self.project.registerData("metadata", pd.DataFrame([{"recordId": str(r)} for r in range(10)]))
        RecordLoader.getRecordList = MagicMock(return_value=["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"])

        self.project.registerData("data-processed", dataProcessed)
        self.project.registerData("data-features", dataFeatures)

    def testDataTraining(self):
        data = self.getData("dataset-training")

        self.assertEqual(len(data), 4)
        data = iter(data)

        x, y = data.__next__()
        npt.assert_equal(x[0, :, 0], np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0]))
        npt.assert_equal(x[0, :, 1], np.array([0, 0, 0, 0, 0, 1, 0, 0, 0, 0]))
        npt.assert_equal(y[0], np.array([-1, -1, -1, -1, -1, 0, -1, -1, -1, -1]).reshape(-1, 1))

        x, y = data.__next__()
        npt.assert_equal(x[0, :, 0], np.array([0, 0, 0, 0, 0, 2, 0, 0, 0, 0]))
        npt.assert_equal(x[0, :, 1], np.array([0, 0, 0, 0, 0, 3, 0, 0, 0, 0]))
        npt.assert_equal(y[0], np.array([-1, -1, -1, -1, -1, 1, -1, -1, -1, -1]).reshape(-1, 1))

        x, y = data.__next__()
        npt.assert_equal(x[0, :, 0], np.array([0, 0, 0, 0, 0, 4, 0, 0, 0, 0]))
        npt.assert_equal(x[0, :, 1], np.array([0, 0, 0, 0, 0, 5, 0, 0, 0, 0]))
        npt.assert_equal(y[0], np.array([-1, -1, -1, -1, -1, 2, -1, -1, -1, -1]).reshape(-1, 1))

        x, y = data.__next__()
        npt.assert_equal(x[0, :, 0], np.array([0, 0, 0, 0, 0, 6, 0, 0, 0, 0]))
        npt.assert_equal(x[0, :, 1], np.array([0, 0, 0, 0, 0, 7, 0, 0, 0, 0]))
        npt.assert_equal(y[0], np.array([-1, -1, -1, -1, -1, 3, -1, -1, -1, -1]).reshape(-1, 1))

    def testDataValidation(self):
        data = self.getData("dataset-validation")

        self.assertEqual(len(data), 3)
        data = iter(data)

        x, y = data.__next__()
        npt.assert_equal(x[0, :, 0], np.array([0, 0, 0, 0, 0, 8, 0, 0, 0, 0]))
        npt.assert_equal(x[0, :, 1], np.array([0, 0, 0, 0, 0, 9, 0, 0, 0, 0]))
        npt.assert_equal(y[0], np.array([-1, -1, -1, -1, -1, 4, -1, -1, -1, -1]).reshape(-1, 1))

        x, y = data.__next__()
        npt.assert_equal(x[0, :, 0], np.array([0, 0, 0, 0, 0, 10, 0, 0, 0, 0]))
        npt.assert_equal(x[0, :, 1], np.array([0, 0, 0, 0, 0, 11, 0, 0, 0, 0]))
        npt.assert_equal(y[0], np.array([-1, -1, -1, -1, -1, 5, -1, -1, -1, -1]).reshape(-1, 1))

        x, y = data.__next__()
        npt.assert_equal(x[0, :, 0], np.array([0, 0, 0, 0, 0, 12, 0, 0, 0, 0]))
        npt.assert_equal(x[0, :, 1], np.array([0, 0, 0, 0, 0, 13, 0, 0, 0, 0]))
        npt.assert_equal(y[0], np.array([-1, -1, -1, -1, -1, 6, -1, -1, -1, -1]).reshape(-1, 1))

    def testDataTest(self):
        data = self.getData("dataset-test")

        self.assertEqual(len(data), 3)
        data = iter(data)

        x, y = data.__next__()
        npt.assert_equal(x[0, :, 0], np.array([0, 0, 0, 0, 0, 14, 0, 0, 0, 0]))
        npt.assert_equal(x[0, :, 1], np.array([0, 0, 0, 0, 0, 15, 0, 0, 0, 0]))
        npt.assert_equal(y[0], np.array([-1, -1, -1, -1, -1, 7, -1, -1, -1, -1]).reshape(-1, 1))

        x, y = data.__next__()
        npt.assert_equal(x[0, :, 0], np.array([0, 0, 0, 0, 0, 16, 0, 0, 0, 0]))
        npt.assert_equal(x[0, :, 1], np.array([0, 0, 0, 0, 0, 17, 0, 0, 0, 0]))
        npt.assert_equal(y[0], np.array([-1, -1, -1, -1, -1, 8, -1, -1, -1, -1]).reshape(-1, 1))

        x, y = data.__next__()
        npt.assert_equal(x[0, :, 0], np.array([0, 0, 0, 0, 0, 18, 0, 0, 0, 0]))
        npt.assert_equal(x[0, :, 1], np.array([0, 0, 0, 0, 0, 19, 0, 0, 0, 0]))
        npt.assert_equal(y[0], np.array([-1, -1, -1, -1, -1, 9, -1, -1, -1, -1]).reshape(-1, 1))


    @OverwriteConfig({"BuildDataset": {"useMultiThreading": True, "threads": 2}})
    def testValidationMultiThreading(self):
        with contextlib.suppress(KeyError):
            self.project.unregister("dataset-validation")
        self.testDataValidation()

    @OverwriteConfig({"BuildDataset": {"useMultiThreading": True, "threads": 2}})
    def testTrainingMultiThreading(self):
        with contextlib.suppress(KeyError):
            self.project.unregister("dataset-training")
        self.testDataTraining()




import pandas as pd
from pyPhases import Phase
from pyPhases.util import BatchProgress
from pyPhasesRecordloader import AnnotationNotFound, RecordLoader, AnnotationInvalid, ParseError
from pyPhasesML import DataversionManager
from SleePyPhases.PreManipulation import PreManipulation

import tqdm


class ExtractEvents(Phase):
    """
    Extract numpy arrays from the raw records in the dataset
    """

    useMultiThreading = True
    threads = 8
    validatDataset = True
    manipulationSteps = None

    def getEventList(self, recordId):
        rl = RecordLoader.get()
        try:
            eventList = rl.getEventList(recordId)

            if self.manipulationSteps is not None:
                dataManipulation = PreManipulation.getInstance(self.manipulationSteps)
                _, eventList = dataManipulation(None, eventList)

            recordDf = pd.DataFrame(eventList)
            recordDf["recordId"] = recordId
            return recordDf
        except AnnotationNotFound:
            return None
        except AnnotationInvalid:
            return None
        except ParseError:
            return None


    def getEvents(self, records, force=False):
        rl = RecordLoader.get()

        bp = BatchProgress(records)
        bp.useMultiThreading = self.useMultiThreading
        if self.threads is not None:
            bp.threads = self.threads
            

        def combine(recordDfs, index):

            # combine.df.append([e for e in events if e is not None])
            df = pd.concat(recordDfs, ignore_index=True)
            combine.df = pd.concat([combine.df, df], ignore_index=True)
            # pd.concat(recordDfs, ignore_index=True)
        # def combineFlatten(events, index):
        #     combine.events += [e for ev in events if ev is not None for e in ev]
        # def getEventList(recordId):
        #     rl.getEventList(recordId)
        combine.df = pd.DataFrame()

        # bp.asyncQueue(rl.getEventList, records, combine)
        bp.start(self.getEventList, records, combine)
        return combine.df
    
    def appendDataframe(self, df, dfToAppend, colMap, offsetStart=None, offsetEnd=None):
        cols = {n: [] for n in colMap}
        if offsetStart is None:
            offsetStart = 0
        if offsetEnd is None:
            offsetEnd = 0
        for e in df.iloc:
            start = e.start + offsetStart
            end = e.end + offsetEnd
            possibleEvents = dfToAppend.query("start < %f and end > %f" % (end, start))
            for name in cols:
                cols[name].append(",".join([ev[colMap[name]] for ev in possibleEvents.iloc]))
        for name in cols:
            df[name] = cols[name]

    def generateData(self, name):
        if name == "events":
            # groupedRecords = self.project.getData("allDBRecordIds", list)
            dm = self.getData("dataversionmanager", DataversionManager)
            groupedRecords = dm.groupedRecords
            
            flatten = [r for records in groupedRecords.values() for r in records]
            # extract the data
            events = self.getEvents(flatten)
            self.registerData("events", events)
        if name == "eventDF":
            # allRecordEventsBatches = self.getData("events", list)
            df = self.getData("events", pd.DataFrame)
            # events = []
            # recordIndex = 0
            # for allRecordEvents in tqdm.tqdm(allRecordEventsBatches):
            #     for recordEvents in allRecordEvents:
            #         for event in recordEvents:
            #             events.append({
            #                 "recordIndex": recordIndex,
            #                 "name": event.name,
            #                 "start": event.start,
            #                 "duration": event.duration,
            #             })
            #         recordIndex += 1
            # df = pd.DataFrame.from_dict(events)
            df["end"] = df["start"] + df["duration"]
            dfSleepStage = df.query("name in ['W', 'N1', 'N2', 'N3', 'R']")
            dfArousal = df.query("name in ('arousal', 'arousal_rera')")
            dfLM = df.query("name in ('LegMovement-Left', 'LegMovement-Right')")
            eventResp = df.query("name in ('resp_hypopnea', 'resp_centralapnea', 'resp_obstructiveapnea', 'resp_mixedapnea', 'resp_cheynestokesbreath')")

            # self.appendDataframe(dfArousal, dfSleepStage, {"sleepStage": "name"}, offsetStart=-0.5, offsetEnd=0.5)
            # self.appendDataframe(dfArousal, dfLM, {"legmovements": "name"}, offsetStart=-0.5, offsetEnd=0.5)
            # self.appendDataframe(dfArousal, eventResp, {"resp_event": "name"}, offsetStart=-0.5, offsetEnd=0.5)

            self.registerData("eventDF", df)
            
            return dfArousal
    
    def appendDataframe(self, df, dfToAppend, colMap, offsetStart=None, offsetEnd=None, idQuery=""):
        cols = {n: [] for n in colMap}
        if offsetStart is None:
            offsetStart = 0
        if offsetEnd is None:
            offsetEnd = 0
        for e in tqdm.tqdm(df.iloc, total=len(df)):
            start = e.start + offsetStart
            end = e.end + offsetEnd
            possibleEvents = dfToAppend.query(f"recordId == '{e.recordId}' and start < {end} and end > {start}")
            for name in cols:
                cols[name].append(",".join([ev[colMap[name]] for ev in possibleEvents.iloc]))
        for name in cols:
            df[name] = cols[name]
            
    def plotArousalHist(self, df, col, name):
        from matplotlib import pyplot as plt
        
        df[col].plot(kind='hist', bins=100, density=False)
        plt.xlabel(name)
        # plt.ylabel('Density')
        plt.ylabel('Arousal Count')
        plt.title('Distribution of %s (M=%.2f, μ=%.2f, σ=%.2f)'%(name, df["duration"].median(), df["duration"].mean(), df["duration"].std()))



    def main(self):
        from matplotlib import pyplot as plt
        import matplotlib.ticker as ticker
        
        from SleePyPhases.Plot import Plot
        
        # df = self.getData("eventDF")
        df = self.getData("eventDF", list)

        df["site"] = df["recordId"].str[:5]
        
        modelConfigString = self.project.getDataFromName("eventDF").getTagString()
        logPath = f"eval/{modelConfigString}/"
        self.logPath = logPath
        plot = Plot(logPath)
        plot.createLogFolder()
        dfArousal = df.query("name.str.contains('arousal')")
        dfLeg = df.query("name.str.contains('Leg')")
        plot.plotClassDistribution(dfArousal.groupby("name")["recordId"].count(), "Arousal/Rera")
        plot.save("arousal-rera")
        # plot.plotClassDistribution(df.query("name == 'arousal'").groupby("sleepStage")["recordIndex"].count(), "sleepStage Arousal")
        # plot.save("arousal-sleepStage")
        # plot.plotClassDistribution(df.query("name == 'arousal_rera'").groupby("sleepStage")["recordIndex"].count(), "sleepStage Rera")
        # plot.save("rara-sleepStage")
        # plot.plotClassDistribution(df.query("name == 'arousal_rera'").groupby("sleepStage")["recordIndex"].count(), "sleepStage")
        # plot.save("sleepStage")
        
        # df["type"] = df.apply(lambda row: ("L" if row["legmovements"] != "" else "") + ("R" if row["resp_event"] != "" else ""), axis=1)
        # df["first_resp"] = df.apply(lambda row: row["resp_event"].split(",")[0], axis=1)
        
        # plot.plotClassDistribution(df.groupby("type")["recordIndex"].count(), "type")
        # plot.save("type")
        
        # plot.plotClassDistribution(df.query("first_resp != ''").groupby("first_resp")["recordIndex"].count())
        # plot.save("first_resp")
        
        
        # plot.plotClassDistribution(df.groupby("resp_event")["recordIndex"].count(), "resp_events")
        # plot.save("resp_events")
        
        
        # df["duration"].plot(kind='density')
        self.plotArousalHist(dfArousal, "duration", "Duration")
        plt.text(10 + 2.5, plt.ylim()[1]*0.9, "10", color='red', ha='center', va='bottom')
        plt.axvline(x=10, color='red')
        plot.save("duration")
        
        self.plotArousalHist(dfArousal.query("name == 'arousal_rera'"), "duration", "Rera Duration")
        plt.text(10 + 2.5, plt.ylim()[1]*0.9, "10", color='red', ha='center', va='bottom')
        plt.axvline(x=10, color='red')
        plot.save("rera-duration")
        
        self.plotArousalHist(dfArousal.query("name == 'arousal'"), "duration", "Arousal Duration")
        plt.text(10 + 2.5, plt.ylim()[1]*0.9, "10", color='red', ha='center', va='bottom')
        plt.axvline(x=10, color='red')
        plot.save("arousal-duration")
        
        self.plotArousalHist(dfArousal.query("name == 'arousal'"), "start", "Arousal Start")
        plot.save("arousal-start")
        
        self.plotArousalHist(dfArousal.query("name == 'arousal_rera'"), "start", "Rera Start")
        plot.save("rera-start")
        
        self.plotArousalHist(dfArousal, "start", "Start")
        plt.gca().xaxis.set_major_formatter(ticker.FuncFormatter(lambda x, pos: f"{x/3600:.1f}h"))
        plot.save("start")
        
        



import numpy as np
import pandas as pd
from pyPhases import Phase
from pyPhasesML.datapipes import Manipulator, DatasetXY
from pyPhasesML import DataversionManager, DataLoader, RecordMap, MultiSourceMap
from pyPhasesML.exporter.MemmapRecordSegmentsExporter import (
    MemmapRecordSegmentsExporter,
)

from SleePyPhases.DataManipulation import DataManipulation
from pyPhasesRecordloader import RecordLoader


class BuildDataset(Phase):
    """
    in this phase the raw data is loaded, manipulated and converted into a dataset that can be used for training
    """

    def buildDataRecords(self):
        splitName = self.getConfig("datasetSplit")

        datasetVersions = self.getConfig("datafold.datasets", {"config": {}})
        datasetVersions = datasetVersions if isinstance(datasetVersions, list) else [datasetVersions]

        seed = self.getConfig("datafold.seed", 2024)
        multiSrcMap = MultiSourceMap(seed=None)
        recordMetadata = []
        for sourceVersion in datasetVersions:
            overwriteConfig = sourceVersion["config"]

            with self.project:
                if "dataversion" in overwriteConfig:
                    self.project.setConfig("dataversion", {})

                self.project.updateConfig(overwriteConfig)
                memmapOptions = {
                    "dtype": self.getConfig("preprocessing.dtype", "float32"),
                }
                dataExporterSignals = self.project.getData("data-processed", np.memmap, options=memmapOptions)
                dataExporterFeatures = self.project.getData("data-features", np.memmap, options=memmapOptions)

                dm = self.getData("dataversionmanager", DataversionManager)
                metadata = self.project.getData("metadata", pd.DataFrame)
                metadataMap = {r["recordId"]: r for r in metadata.iloc}
                allRecordIds = [r for subgroup in dm.unShuffledgroupedRecords.values() for r in subgroup]

                recordIds = dm.getRecordsForSplit(splitName)
                recordIndexes = [allRecordIds.index(r) for r in recordIds]
                recordMetadata = recordMetadata + [metadataMap[r] for r in recordIds]

                # handle segment wise data
                recordLengthsInSegments = None
                if not self.getConfig("recordWise"):
                    segmentLength = self.getConfig("segmentLength")
                    segmentLengthLabel = self.getConfig("segmentLengthLabel", segmentLength)
                    paddedSegments = self.getConfig("segmentPadding", [0, 0])
                    recordLengths = dataExporterSignals.recordLengths

                    dataExporterSignals = MemmapRecordSegmentsExporter(dataExporterSignals, segmentLength, paddedSegments)
                    dataExporterFeatures = MemmapRecordSegmentsExporter(
                        dataExporterFeatures, segmentLengthLabel, paddedSegments
                    )

                    recordLengthsInSegments = [length // segmentLength - sum(paddedSegments) for length in recordLengths]
                    assert all(np.array(recordLengthsInSegments) > 0), (
                        "segmentLength (including segmentPadding) is too large for some recordings"
                    )

                dataset = DatasetXY(dataExporterSignals, dataExporterFeatures)
                recordSet = RecordMap(dataset, recordIndexes, mappingLengths=recordLengthsInSegments)

                multiSrcMap.addRecordMap(recordSet)

        # shuffle recordings or segments
        if self.getConfig("datafold.shuffle", False):
            multiSrcMap.shuffle()
            # use same seed to shuffle the metadata
            np.random.seed(seed)
            np.random.shuffle(recordMetadata)

        return multiSrcMap, recordMetadata

    def buildDatasetManipulatorBySplit(self):
        split = self.getConfig("datasetSplit", "training")

        dataRecords, recordMetaData = self.buildDataRecords()

        useConfigName = "segmentManipulationEval" if split != "training" else "segmentManipulation"
        manipulationSteps = self.getConfig(useConfigName)
        segmentManipulation = DataManipulation.getInstance(
            manipulationSteps, split, self.project.config, recordMetadata=recordMetaData
        )
        return Manipulator(dataRecords, segmentManipulation)

    def buildDatasetBySplit(self):
        split = self.getConfig("datasetSplit")
        batchSize = self.getConfig("trainingParameter.batchSize") if split != "test" else self.getConfig("eval.batchSize")
        buildConfig = self.getConfig("BuildDataset", {"useMultiThreading": True})

        manipulator = self.buildDatasetManipulatorBySplit()

        dataloader = DataLoader.build(
            manipulator,
            preload=buildConfig["useMultiThreading"],
            batchSize=batchSize,
            shuffle=self.getConfig("trainingParameter.shuffle", False),
            shuffleSeed=self.getConfig("trainingParameter.shuffleSeed", 2),
        )

        manipulationSteps = self.getConfig("batchManipulation")
        segmentManipulation = DataManipulation.getInstance(manipulationSteps, split, self.project.config)
        batchManipulation = Manipulator(dataloader, segmentManipulation)

        return batchManipulation

    def generateData(self, name):
        # get a single record based on the config value "recordId"
        if name == "record-data":
            recordId = self.getConfig("recordId")
            recordMetaData = RecordLoader.get().getMetaData(recordId)
            preprocesedRecord = self.getData("record-preprocessed", list)
            split = self.getConfig("datasetSplit", "training")
            useConfigName = "segmentManipulationEval" if split != "training" else "segmentManipulation"
            manipulationSteps = self.getConfig(useConfigName)

            segmentManipulation = DataManipulation.getInstance(
                manipulationSteps, split, self.project.config, recordMetadata=recordMetaData
            )
            return Manipulator(DatasetXY([preprocesedRecord[0]], [preprocesedRecord[1]]), segmentManipulation)
        splitName = name[8:]
        if splitName != "":
            self.setConfig("datasetSplit", splitName)

        # dont register the data
        return self.buildDatasetBySplit()

    def main(self):
        self.logError(
            "No dataset is build on default. Please use Extract the data for the dataset or Training to train a model."
        )



from pyPhases import Phase
from glob import glob
import os
import re
import pandas as pd
from pathlib import Path
import json
import sys
from pyPhasesML import Scorer


class LogCleanup(Phase):
    def loadConfig(self, json_path):
        with open(json_path, "r") as file:
            config = json.load(file)
        return config

    def parse_checkpoint_filename(self, filename):
        """Parse checkpoint filename to extract epoch and metrics."""
        # Extract base filename without path
        base_filename = os.path.basename(filename)
        
        # Regular expression to match the checkpoint pattern
        pattern = r"checkpointModel_(\d+)_(.+)\.pkl"
        match = re.match(pattern, base_filename)
        
        if not match:
            return None
        
        epoch = int(match.group(1))
        metrics_str = match.group(2)
        metrics = [float(m) for m in metrics_str.split('_')]
        
        return {
            'epoch': epoch,
            'metrics': metrics,
            'filename': filename
        }

    def get_metric_directions(self, log_folder):
        """Get the direction (bigger is better or not) for each metric."""
        # Try to load config to get metric names
        config_path = os.path.join(log_folder, "project.config")
        metric_directions = []
        
        try:
            if os.path.exists(config_path):
                config = self.loadConfig(config_path)
                if "trainingParameter" in config and "validationMetrics" in config["trainingParameter"]:
                    metrics = config["trainingParameter"]["validationMetrics"]
                    # Flatten metrics if they're nested
                    flat_metrics = []
                    for m in metrics:
                        if isinstance(m, list):
                            flat_metrics.extend(m)
                        else:
                            flat_metrics.append(m)
                    
                    # Get direction for each metric
                    scorer = Scorer()
                    for metric in flat_metrics:
                        _, _, bigger_is_better = scorer.getMetricDefinition(metric)
                        metric_directions.append(bigger_is_better)
        except Exception as e:
            print(f"  Warning: Could not determine metric directions from config: {e}")
        
        # If we couldn't determine directions, assume all metrics are "bigger is better"
        if not metric_directions:
            # Try to infer from checkpoint filenames
            checkpoint_files = glob(os.path.join(log_folder, "checkpointModel_*.pkl"))
            if checkpoint_files:
                first_checkpoint = self.parse_checkpoint_filename(checkpoint_files[0])
                if first_checkpoint:
                    metric_count = len(first_checkpoint['metrics'])
                    metric_directions = [True] * metric_count
        
        return metric_directions

    def is_dominated(self, checkpoint, other_checkpoints, metric_directions):
        """Check if a checkpoint is dominated by any other checkpoint, considering metric directions."""
        if not metric_directions:
            # Default to all metrics being "bigger is better"
            metric_directions = [True] * len(checkpoint['metrics'])
        
        for other in other_checkpoints:
            if other['filename'] == checkpoint['filename']:
                continue
            
            # Check if all metrics in other checkpoint are better than or equal to this checkpoint
            all_metrics_better_or_equal = True
            
            for i, (checkpoint_metric, other_metric) in enumerate(zip(checkpoint['metrics'], other['metrics'])):
                bigger_is_better = metric_directions[i] if i < len(metric_directions) else True
                
                if bigger_is_better:
                    if other_metric < checkpoint_metric:
                        all_metrics_better_or_equal = False
                        break
                else:  # Smaller is better
                    if other_metric > checkpoint_metric:
                        all_metrics_better_or_equal = False
                        break
            
            # If other checkpoint has higher epoch and all metrics are better or equal, this checkpoint is dominated
            if other['epoch'] > checkpoint['epoch'] and all_metrics_better_or_equal:
                return True
                
            # Check if all metrics are strictly better regardless of epoch
            all_metrics_strictly_better = True
            for i, (checkpoint_metric, other_metric) in enumerate(zip(checkpoint['metrics'], other['metrics'])):
                bigger_is_better = metric_directions[i] if i < len(metric_directions) else True
                
                if bigger_is_better:
                    if other_metric <= checkpoint_metric:
                        all_metrics_strictly_better = False
                        break
                else:  # Smaller is better
                    if other_metric >= checkpoint_metric:
                        all_metrics_strictly_better = False
                        break
            
            if all_metrics_strictly_better:
                return True
                
        return False

    def main(self):
        log_directory = "logs/"
        
        # Check if "all" option is provided
        confirm_all = "all" in self.getConfig("cleanup.options", [])
        
        # Get all log folders
        log_folders = glob(log_directory + "/*")
        
        for folder in log_folders:
            if not os.path.isdir(folder):
                continue
                
            print(f"\nAnalyzing log folder: {folder}")
            
            # Get all checkpoint files
            checkpoint_files = glob(os.path.join(folder, "checkpointModel_*.pkl"))
            resume_files = glob(os.path.join(folder, ".resume*"))
            
            # Parse checkpoint information
            checkpoints = []
            for file in checkpoint_files:
                parsed = self.parse_checkpoint_filename(file)
                if parsed:
                    checkpoints.append(parsed)
            
            # If no checkpoints found, check if folder should be deleted
            if not checkpoints:
                all_files = os.listdir(folder)
                if len(all_files) <= 2 and all(f in all_files for f in ["project.config", "model.config"]):
                    print(f"  Folder {folder} contains only config files and no checkpoints. Deleting folder.")
                    for file in all_files:
                        os.remove(os.path.join(folder, file))
                    os.rmdir(folder)
                    print(f"  Deleted folder {folder}")
                    continue
            
            # Get metric directions (bigger is better or not)
            metric_directions = self.get_metric_directions(folder)
            
            # Identify obsolete checkpoints
            obsolete_checkpoints = []
            keep_checkpoints = []
            
            for checkpoint in checkpoints:
                if self.is_dominated(checkpoint, checkpoints, metric_directions):
                    obsolete_checkpoints.append(checkpoint)
                else:
                    keep_checkpoints.append(checkpoint)
            
            # Print summary and ask for confirmation if needed
            if obsolete_checkpoints or resume_files:
                print(f"  Found {len(obsolete_checkpoints)} obsolete checkpoints and {len(resume_files)} resume files.")
                print("  Checkpoints to keep:")
                for cp in keep_checkpoints:
                    print(f"    - Epoch {cp['epoch']}, Metrics: {cp['metrics']}")
                
                print("  Checkpoints to delete:")
                for cp in obsolete_checkpoints:
                    print(f"    - Epoch {cp['epoch']}, Metrics: {cp['metrics']}")
                
                if resume_files:
                    print("  Resume files to delete:")
                    for file in resume_files:
                        print(f"    - {os.path.basename(file)}")
                
                proceed = confirm_all
                if not proceed:
                    confirm = input("  Proceed with deletion? (y/n/all): ")
                    proceed = confirm.lower() == 'y' or confirm.lower() == 'all'
                    confirm_all = confirm.lower() == 'all'
                
                if proceed:
                    # Delete obsolete checkpoints
                    for cp in obsolete_checkpoints:
                        os.remove(cp['filename'])
                        print(f"  Deleted: {os.path.basename(cp['filename'])}")
                    
                    # Delete resume files
                    for file in resume_files:
                        os.remove(file)
                        print(f"  Deleted: {os.path.basename(file)}")
                else:
                    print("  Deletion cancelled.")
            else:
                print("  No obsolete checkpoints or resume files found.")
                
        print("\nLog cleanup completed.")



import numpy as np

from pyPhases import Phase
from pyPhasesRecordloader import RecordSignal, Signal

from SleePyPhases.Plot import Plot
from SleePyPhases.DataManipulation import DataManipulation
from datetime import timedelta



class EvalPlotExamples(Phase):
    allowTraining: bool = True
    exampleSkipClass = ["0-0"]
    examplePaddingInS = 10
    highlight = None
    exampleSecondsPerInch = 2

    def getExternalTestset(self, split="test"):
        with self.project:
            evalConfig = self.getConfig("evalOn")

            # try not to update dataversion, rather replace it completly
            if "dataversion" in evalConfig:
                self.project.config["dataversion"] = {}

            self.project.config.update(evalConfig)
            self.project.trigger("configChanged", None)
            dataset = self.project.getData(f"dataset-{split}", list)
            dm = self.getData("dataversionmanager")
            recordsMap = dm.getRecordsForSplit(split)

        return dataset, dm, recordsMap


    def plotRecordSignal(
        self, recordSignal: RecordSignal, fileName, sliceValues=None, highlights=None, offset=0, title="Example"
    ):
        from matplotlib import pyplot as plt
        samplingrate = recordSignal.targetFrequency
        paddingInS = self.examplePaddingInS
        padding = paddingInS * samplingrate
        highlights = []

        sliceValues = sliceValues[0] + offset, sliceValues[1] + offset

        sliceValues = sliceValues[0], sliceValues[0] + 50 * 100

        if self.highlight is not None:
            s, e = self.highlight
            s += paddingInS
            e = e + paddingInS if e > 0 else e - paddingInS
            highlights.append((s, e))

        plt.clf()
        plot = Plot(self.evalPath)
        plot.eventMap = {}
        
        labelNames = self.getConfig("classification.labelNames")
        def mergePrediction(signal, axs, index):

            if signal.name in labelNames:
                fig = plt.gcf()
                fig.delaxes(axs[index+1])

                ax = axs[index]
                ax.set_ylim(0, 1.1)
                axs[index] = ax
                axs[index+1] = ax

                current_height = ax.get_position().height

                for i in range(index+2, len(axs)):
                    axs[i].set_position(axs[i].get_position().translated(0, current_height + 0.02))


        plot.on("beforePlotSignal", mergePrediction)

        thresholds = self.threshold
        def addThreshold(signal, axs, index):
            # from matplotlib import pyplot as plt
            if signal.name in labelNames:
                threshold = thresholds[labelNames.index(signal.name)]
                axs[index].axhline(y=threshold, color="red")
                axs[index].text(1, threshold + 0.02, f"{signal.name} Threshold", color="red")

        plot.on("afterPlotSignal", addThreshold)

        fig = plot.plotRecordSignal(
            recordSignal,
            sliceValues,
            resample=False,
            padding=padding,
            secondsPerInch=self.exampleSecondsPerInch,
            highlights=highlights,
            title=title,
        )

        fig.axes[-1].set_xlabel("Time (s)")
        fig.axes[-1].xaxis.set_major_formatter(plt.FuncFormatter(lambda val, pos: str(timedelta(seconds=val))))

        plot.save(f"examples/{fileName}")


    def plotRecordSignalWithClassExamples(
        self, label, classExamples, recordSignal: RecordSignal, fileName, sliceValues=None, highlights=None, offset=0, classNames=None, threshold=None, labelName=""
    ):

        classificationNames = {}
        for i, name1 in enumerate(classNames):
            for j, name2 in enumerate(classNames):
                classificationNames[f"{i}-{j}"] = f"{name1} as {name2}"

        for classIndex, examples in classExamples.items():
            for eIndex, sliceValues in enumerate(examples):
                className = classificationNames["-".join(classIndex.split("-")[1:])]
                if classIndex in self.exampleSkipClass:
                    continue

                self.plotRecordSignal(
                    recordSignal, 
                    fileName=f"{eIndex}-{fileName}-{label}-{className}", 
                    sliceValues=sliceValues, 
                    highlights=highlights, 
                    offset=offset, 
                    title="Example for %s: %s in record '{record}' from {from}" % (label, className)
                )


    def getRecordSignal(self, recordStats, segmentStats, recordId):
        channelNames = self.getConfig("evalChannelNames")

        exampleCount = 3
        # pick random records
        randomRecordIndexes = np.random.choice(len(recordStats), exampleCount, replace=False)

        testData, dm, recordsMap = self.getExternalTestset()
        recordIds = dm.getRecordsForSplit("test")

        self.setConfig("BuildDataset.useMultiThreading", False)
        da = DataManipulation.getInstance([], "test", self.project.config)

        recordIndex = recordIds.index(recordId)
        samplingrate = self.getConfig("preprocessing.targetFrequency")
        X, Y = testData[recordIndex]

        if not isinstance(X, np.ndarray):
            X = X.numpy()

        if not isinstance(Y, np.ndarray):
            Y = Y.numpy()
            

        recordSignal = RecordSignal.fromArray(
            X, sourceFrequency=samplingrate, targetFrequency=samplingrate, names=channelNames, transpose=True
        )

        # add labels

        recordSignal.recordId = recordIds[recordIndex]
        
        labelNames = self.getConfig("classification.labelNames")
        labelOrdered = {i:label for i, label in enumerate(labelNames)}
        
        for labelIndex, label in labelOrdered.items():
            labelSignal = Signal(label, Y[:, labelIndex], samplingrate)

            recordSignal.addSignal(labelSignal)


            prediction = segmentStats[label][recordIndex]["prediction"]
            # add prediction
            offset = (len(Y[:, 0]) - prediction.shape[0]) // 2
            prediction, _ = da._fixeSizeSingleChannel(prediction.reshape(1, -1, 1), 2**21, fillValue=0, position="center")
            predictionSignal = Signal(f"{label} Prediction", prediction.reshape(-1), samplingrate)
            recordSignal.addSignal(predictionSignal)

        return recordSignal

    def plotExamples(self, recordStats, segmentStats, name="", labelName=""):
        channelNames = self.getConfig("evalChannelNames")

        exampleCount = 3
        # pick random records
        randomRecordIndexes = np.random.choice(len(recordStats), exampleCount, replace=False)

        testData, dm, recordsMap = self.getExternalTestset()
        recordIds = dm.getRecordsForSplit("test")

        self.setConfig("BuildDataset.useMultiThreading", False)
        da = DataManipulation.getInstance([], "test", self.project.config)

        for i, recordIndex in enumerate(randomRecordIndexes):
            # create recordsignal for processed data
            samplingrate = self.getConfig("preprocessing.targetFrequency")
            X, Y = testData[recordIndex]
            classExamples = recordStats[recordIndex][f"examples_{labelName}"]

            if not isinstance(X, np.ndarray):
                X = X.numpy()

            if not isinstance(Y, np.ndarray):
                Y = Y.numpy()
                

            recordSignal = RecordSignal.fromArray(
                X, sourceFrequency=samplingrate, targetFrequency=samplingrate, names=channelNames, transpose=True
            )

            # add labels

            recordSignal.recordId = recordIds[recordIndex]
            
            labelNames = self.getConfig("classification.classNames")
            mainLabelName = labelName
            mainLabelIndex = labelNames.index(labelName)

            # labelOrdered = {i:label for i, label in enumerate(labelNames) if label != mainLabelName}
            # labelOrdered[mainLabelIndex] = mainLabelName
            labelOrdered = {i:label for i, label in enumerate(labelNames)}
            # labelOrdered[mainLabelIndex] = mainLabelName


            
            for labelIndex, label in labelOrdered.items():
                labelSignal = Signal(label, Y[:, labelIndex], samplingrate)
                # labelSignal = Signal(classificationName, Y[:, 0], samplingrate)
                recordSignal.addSignal(labelSignal)


                prediction = segmentStats[label][recordIndex]["prediction"]
                # add prediction
                offset = (len(Y[:, 0]) - prediction.shape[0]) // 2
                prediction, _ = da._fixeSizeSingleChannel(prediction.reshape(1, -1, 1), 2**21, fillValue=0, position="center")
                predictionSignal = Signal(f"{label} Prediction", prediction.reshape(-1), samplingrate)
                recordSignal.addSignal(predictionSignal)

            classNames = [f"No {labelName}", labelName]
            classExamples = {index: [(c[0]+offset, c[1]+offset) for c in ce] for index, ce in classExamples.items()}
            self.plotRecordSignalWithClassExamples(labelName, classExamples, recordSignal, f"example-{i}-processed", classNames=classNames, offset=offset, threshold=self.threshold[mainLabelIndex], labelName=labelName)

   
    def plotVideoExample(self, recordStats, segmentStats, recordId, start, end):
        
        testData, dm, recordsMap = self.getExternalTestset()
        recordIds = dm.getRecordsForSplit("test")

        recordId = "acq_238127226"
        recordSignal = self.getRecordSignal(recordStats, segmentStats, recordId)

        start = 3*60*60 + 18 + 60 + 34
        start *= 50
        end = start + 5*60*50
        end = 2097152

        start = 648600
        end = start + 5*60*50
        start, end = (618600, 678600)
        # start, end = (510000, 550000)

        export = self.project.getExporterForType(list)
        export.write("myexample", recordSignal)
        recordSignal = export.read("myexample")

        self.plotRecordSignal(
            recordSignal=recordSignal,
            fileName="blub",
            sliceValues=[start,end]
        )
        # self.plotRecordSignal("Sleep", classExamples, recordSignal, f"example-{i}-processed", classNames=classNames, offset=offset, threshold=self.threshold[mainLabelIndex], labelName=labelName)
        pass



    def main(self):
        import pandas as pd
        
        threshold = self.getConfig("fixedThreshold", False)
        self.threshold = threshold or self.getData("threshold", float)

        modelConfigString = self.project.getDataFromName("eventResults").getTagString()
        evalPath = f"eval/{modelConfigString}/"
        self.evalPath = evalPath
        labelNames = self.getConfig("classification.labelNames")
        segmentResultsRecords, segmentResults = self.getData("evalResults", list)
        resultRows, result = self.getData("eventResults", list)

        start = 3*60*60 + 18 + 60 + 34
        start *= 50
        end = start + 5*60*50

        # self.plotVideoExample(resultRows, segmentResultsRecords, "acq_238127226", start, end)
        for label in labelNames:
            self.plotExamples(recordStats=resultRows, segmentStats=segmentResultsRecords, labelName=label)
        self.log(f"Eval finished: {evalPath}")




import numpy as np
from pyPhases import Phase
from pyPhasesRecordloader import ChannelsNotPresent, RecordLoader
from tqdm import tqdm


class ExtractRecordFeatures(Phase):
    """
    load record ids
    """

    def addEEGSignalQuality(self, signal):
        from scipy.stats import kurtosis, skew
        if signal.dimension in ["uV"]:
            pass
        elif signal.dimension in ["mV"]:
            signal.signal = signal.signal * 1000
        else:
            self.logError(f"Unkown Signaldimension '{str(signal.dimension)}' in signal '{signal.name}', assuming 'uV'")
        signal.bandpass(low=0.3, high=35, order=3)
        defaultWindowSize = 8
        if signal.signal.shape[0] % (defaultWindowSize * signal.frequency):
            signal.signal = signal.signal[
                : int(signal.signal.shape[0] / (defaultWindowSize * signal.frequency))
                * int(defaultWindowSize * signal.frequency)
            ]
        windowed = np.reshape(
            signal.signal,
            (round(len(signal.signal) / (defaultWindowSize * signal.frequency)), round(defaultWindowSize * signal.frequency)),
        )
        # Kriterien inkl. Schwelllwerten aus der Veröffentlichung DOI: 10.1109/JBHI.2019.2920381
        highAmpl = np.sum(abs(windowed) > 151.09, axis=1) / windowed.shape[1] > (
            1 / defaultWindowSize * 0.5
        )  # 95: 95, 100: 151.09
        highStd = np.std(windowed, axis=1) > 33.23  # 95: 22.29, 100: 33.23
        # highApEn = ... > 0.65 # 95: 1.01, 100: 0.65
        # highAmplVar = (np.max(abs(windowed), axis=1) / np.square(np.std(windowed, axis=1))) > 1.02e5 # 95: 1.56e5, 100: 1.02e5
        highKurt = kurtosis(windowed, axis=1, bias=False) > 14.56  # 95: 6.53, 100: 14.56
        highSkew = abs(skew(windowed, axis=1, bias=False)) > 1.79  # 95: 0.69, 100: 1.79

        sumArtefacts = highAmpl + highStd + highKurt + highSkew
        signal.quality = 1 - (sum(sumArtefacts) / len(sumArtefacts))

        return signal.quality

    def main(self):
        metaData = self.project.getData("metadata", list)
        rl = RecordLoader.get()
        for record in tqdm(metaData):
            recordId = record["recordId"]
            if "EEG F4-A1-quality" in record or not record["annotationExist"]:
                continue
            try:
                recordSignal = rl.getSignal(recordId)
            except ChannelsNotPresent:
                continue

            for channel in recordSignal.signals:
                if channel.typeStr == "eeg":
                    quality = self.addEEGSignalQuality(channel)
                    record[channel.name + "-quality"] = quality
        self.registerData("metadata", metaData)



import unittest

import numpy as np
import numpy.testing as npt
from numpy.random import default_rng
from pyPhases import classLogger

from SleePyPhases.DataManipulation import DataManipulation


@classLogger
class TestDataManipulation(unittest.TestCase):
    def getDA(self, numLabels=2):
        return DataManipulation.getInstance(
            [],
            "Test",
            {
                "numLabels": numLabels,
                "numClasses": [[2]],
            },
            seed=2,
        )

    def test_combineWithPatience(self):
        da = self.getDA()

        prediction = np.array([0, 1, 0, 0, 1, 0, 1, 0, 0, 1, 1])
        expected = np.array([0, 1, 0, 0, 1, 1, 1, 0, 0, 1, 1])
        patience = 2
        X = da._combineWithPatience(prediction, patience)

        npt.assert_equal(X, expected)
        
    def test_combineWithPatienceMoreSegments(self):
        da = self.getDA()

        prediction = np.array([0, 1, 0, 0, 1, 0, 1, 0, 1, 0, 1, 1])
        expected = np.array([0, 1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1])

        X = da._combineWithPatience(prediction, patience=2)

        npt.assert_equal(X, expected)

    def test_combineWithPatienceMinority2(self):
        da = self.getDA()

        prediction = np.array([0, 1, 0, 0, 0, 1])
        expected = np.array([0, 1, 1, 1, 1, 1])

        X = da._combineWithPatience(prediction, patience=4)

        npt.assert_equal(X, expected)

    def test_combineWithPatienceMulticlassGap(self):
        da = self.getDA()

        prediction = np.array([0, 1, 0, 0, 1, 2, 1, 0, 0, 1, 1])
        expected = np.array([0, 1, 0, 0, 1, 1, 1, 0, 0, 1, 1])
        patience = 2
        X = da._combineWithPatience(prediction, patience)

        npt.assert_equal(X, expected)
        
    def test_combineWithPatienceMulticlassEnd(self):
        da = self.getDA()

        prediction = np.array([0, 1, 0, 0, 2, 2, 1, 0, 0, 1, 1])
        expected = np.array([0, 1, 0, 0, 2, 2, 2, 0, 0, 1, 1])
        patience = 2
        X = da._combineWithPatience(prediction, patience)

        npt.assert_equal(X, expected)
        
    def test_combineWithPatienceMulticlassStart(self):
        da = self.getDA()

        prediction = np.array([0, 1, 0, 0, 1, 2, 2, 0, 0, 1, 1])
        expected = np.array([0, 1, 0, 0, 2, 2, 2, 0, 0, 1, 1])
        patience = 2
        X = da._combineWithPatience(prediction, patience)

        npt.assert_equal(X, expected)

    def test_minLength(self):
        da = self.getDA()

        prediction = np.array([0, 1, 1, 0, 0, 1, 0, 1, 0, 0, 0, 1])
        expected = np.array([0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0])

        Y = da._minLength(prediction, minLength=2)

        npt.assert_equal(Y, expected)

    def test_toArousal(self):
        da = self.getDA()
        da.threshold = 0.5

        prediction = np.array([0, 1, 1, 0, 0, 1, 0, 1, 0, 1, 0, 0]).reshape(1, -1, 1)
        expected = np.array([0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0]).reshape(1, -1, 1)

        X, _ = da.toArousal(prediction, None, patience=2, minLength=3, frequency=1)

        npt.assert_equal(X, expected)

    def test_hotEncode(self):
        da = self.getDA(numLabels=3)
        Y = np.array([[0, 1, 2], [1, 0, 2]])
        expected_output = np.array([[[1, 0, 0], [0, 1, 0], [0, 0, 1]], [[0, 1, 0], [1, 0, 0], [0, 0, 1]]])
        _, Y_encoded = da.hotEncode(None, Y)
        self.assertTrue(np.array_equal(Y_encoded, expected_output))

    def test_znorm(self):
        X = np.array([[1, 2, 3, 4], [5, 6, 7, 8]])
        X, _ = self.getDA().znorm(X, None)
        expected_output = np.array(
            [[-1.34164079, -0.4472136, 0.4472136, 1.34164079], [-1.34164079, -0.4472136, 0.4472136, 1.34164079]]
        )
        self.assertTrue(np.allclose(X, expected_output))

    def test_MagScale(self):
        X = np.array([[1, 2, 3, 4], [5, 6, 7, 8]], dtype=np.float32)
        Xscaled, _ = self.getDA().MagScale(X, None)
        rng = default_rng(seed=2)
        scale = 0.8 + rng.random(1, dtype=np.float32) * 0.45
        self.assertTrue(np.allclose(Xscaled, X * scale))

    def test_channelShuffle(self):
        da = self.getDA()
        X = np.array([[1, 2, 3, 4], [5, 6, 7, 8]]).reshape(1, -1, 2)
        XShuffled, _ = da.channelShuffle(X, None, (1, 3))
        self.assertTrue(np.array_equal(XShuffled, X))
        XShuffled, _ = da.channelShuffle(X, None, (1, 3))
        XShuffled, _ = self.getDA().channelShuffle(X, None, (1, 3))

    def test_fixeSizeSingleChannel(self):
        new_size = 7
        fill_value = 1
        position = "center"
        X = np.array(
            [
                [[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12], [13, 14, 15]],
                [[16, 17, 18], [19, 20, 21], [22, 23, 24], [25, 26, 27], [28, 29, 30]],
            ]
        )
        newX, _ = self.getDA()._fixeSizeSingleChannel(X, new_size, fill_value, position)

        self.assertEqual(newX.shape, (2, 7, 3))
        self.assertTrue(np.allclose(newX[:, 1:-1, :], X))
        self.assertTrue(np.allclose(newX[:, 0, :], newX[:, -1, :]))
        self.assertTrue(np.allclose(newX[:, 0, :], fill_value))

    def setUp(self):
        self.test_X = np.arange(600).reshape((10, 20, 3))
        self.test_Y = np.arange(300).reshape((10, 30, 1))
        self.test_factor = 2
        self.test_channels = [0, 2, 4]

    def test_restoreLength(self):
        length = 16
        _, new_Y = self.getDA().restoreLength(self.test_X, self.test_Y, length)
        self.assertEqual(new_Y.shape[0], self.test_Y.shape[0])
        self.assertEqual(new_Y.shape[1], length)
        self.assertEqual(new_Y.shape[2], self.test_Y.shape[2])
        npt.assert_equal(new_Y, self.test_Y[:, 7:23, :])

    def test_selectChannel(self):
        channel = 1
        new_X, new_Y = self.getDA().selectChannel(self.test_X, self.test_Y, channel)
        self.assertEqual(new_X.shape, (10, 20, 1))
        self.assertEqual(new_Y.shape, (10, 30, 1))
        self.assertTrue(np.array_equal(new_X[:, :, 0], self.test_X[:, :, channel]))
        self.assertTrue(np.array_equal(new_Y, self.test_Y))

    def test_selectChannelRandom(self):
        selectChannels = [0, 1, 3]
        new_X, new_Y = self.getDA().selectChannelRandom(self.test_X, self.test_Y, selectChannels)
        self.assertEqual(new_X.shape, (10, 20, 1))
        self.assertEqual(new_Y.shape, (10, 30, 1))
        self.assertTrue(np.array_equal(new_X, self.test_X[:, :, 2:3]))
        self.assertTrue(np.array_equal(new_Y, self.test_Y))

    def test_changeType(self):
        dtype = np.float64
        new_X, new_Y = self.getDA().changeType(self.test_X, self.test_Y, dtype)
        self.assertEqual(new_X.dtype, dtype)
        self.assertEqual(new_Y.dtype, dtype)
        self.assertTrue(np.array_equal(new_X, self.test_X.astype(dtype)))
        self.assertTrue(np.array_equal(new_Y, self.test_Y.astype(dtype)))

    def test_reduceY(self):
        reductionFactor = 3
        prediction = np.array([0.2, 0.3, 0.5, 0.4, 0.6, 0.7, 0.8, 0.9, 0.1]).reshape(1, -1, 1)
        Y = np.array([0, 0, 1, 1, 1, 1, 0, 1, 1]).reshape(1, -1, 1)
        
        new_X, new_Y = self.getDA().reduceY(prediction, Y, reductionFactor)
        npt.assert_almost_equal(new_X.reshape(-1), np.array([1/3, 1.7/3, 1.8/3]))
        npt.assert_equal(new_Y.reshape(-1), np.array([0, 1, 1]))

    def test_reduce_max(self):
        reductionFactor = 3
        prediction = np.array([0.2, 0.3, 0.5, 0.4, 0.6, 0.7, 0.8, 0.9, 0.1, 0.8, 0.9, 0.1]).reshape(1, -1, 1)
        Y = np.array([0, 0, 1, 1, 1, 1, 0, 1, 1, 0, 0, 0]).reshape(1, -1, 1)
        
        new_X, new_Y = self.getDA().reduceY(prediction, Y, reductionFactor, reduce="max")
        npt.assert_almost_equal(new_X.reshape(-1), np.array([0.5, 0.7, 0.9, 0.9]))
        npt.assert_equal(new_Y.reshape(-1), np.array([1, 1, 1, 0]))

    def test_deleteIgnored(self):
        testY = np.array([1, 2, 3, -1, 2, 2, 3, -1, 2, -1, 2]).reshape(1, -1, 1)
        testX = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]).reshape(1, -1, 1)
        newX, newY = self.getDA().deleteIgnored(testX, testY)
        npt.assert_equal(newY.reshape(-1), [1, 2, 3, 2, 2, 3, 2, 2])
        npt.assert_equal(newX.reshape(-1), [1, 2, 3, 5, 6, 7, 9, 11])

    def test_derive_basic(self):
        X = np.arange(1*2*10).reshape(1, 2, 10).transpose(0, 2, 1)
        Y = np.array([0, 1])
        channels = [1, 0]
        X_result, Y_result = self.getDA().derive(X, Y, channels)
        X_expected = np.arange(1*2*10).reshape(1, 2, 10)
        X_expected = np.array([[
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            [10, 11, 12, 13, 14, 15, 16, 17, 18, 19],
            [10, 10, 10, 10, 10, 10, 10, 10, 10, 10]
        ]]).transpose(0, 2, 1)
        self.assertTrue(np.array_equal(X_result, X_expected))
        self.assertTrue(np.array_equal(Y_result, Y))

    def test_extendEvents(self):
        da = self.getDA()
        
        # Create test data with a single event
        Y = np.zeros((1, 10, 1))
        Y[0, 4:6, 0] = 1  # Event from index 4 to 5
        X = np.ones_like(Y)  # Dummy X data
        
        # Test extending events by 2 samples before and 1 after
        X_new, Y_new = da.extendEvents(X, Y, channel=0, pre_samples=2, post_samples=1)
        
        # Expected result: Event should now span from index 2 to 6
        expected = np.zeros((1, 10, 1))
        expected[0, 2:7, 0] = 1
        
        npt.assert_equal(Y_new.shape, expected.shape)
        npt.assert_equal(Y_new[0, :, 0], expected[0, :, 0])
        npt.assert_equal(Y_new, expected)
        npt.assert_equal(X_new, X)  # X should remain unchanged

    def test_extendEventsWithIgnoreValues(self):
        da = self.getDA()
        
        # Create test data with a single event
        Y = np.zeros((1, 10, 1))
        Y[0, 0, 0] = -1
        Y[0, -1, 0] = -1

        Y[0, 4:6, 0] = 1  # Event from index 4 to 5
        X = np.ones_like(Y)  # Dummy X data
        
        # Test extending events by 2 samples before and 1 after
        X_new, Y_new = da.extendEvents(X, Y, channel=0, pre_samples=2, post_samples=1)
        
        # Expected result: Event should now span from index 2 to 6
        expected = np.zeros((1, 10, 1))
        expected[0, 0, 0] = -1
        expected[0, -1, 0] = -1
        expected[0, 2:7, 0] = 1
        
        npt.assert_equal(Y_new.shape, expected.shape)
        npt.assert_equal(Y_new[0, :, 0], expected[0, :, 0])
        npt.assert_equal(X_new, X)  # X should remain unchanged
        
    def test_combineY(self):
        da = self.getDA()
        
        X = np.zeros((2, 10, 2))
        Y = np.array([
            [[0], [1], [2], [3], [4], [0], [1], [2], [3], [4]],
            [[4], [3], [2], [1], [0], [4], [3], [2], [1], [0]]
        ])

        # Test combining values 2,1 into 0
        _, Y_out = da.combineY(X, Y.copy(), values=[2, 1, 0])
        expected = np.array([
            [[0], [0], [0], [1], [2], [0], [0], [0], [1], [2]],
            [[2], [1], [0], [0], [0], [2], [1], [0], [0], [0]]
        ])

        np.testing.assert_array_equal(Y_out, expected)
        
        # Test combining values 4,3,2 into 1
        _, Y_out = da.combineY(X, Y.copy(), values=[1, 2])
        expected = np.array([
            [[0], [1], [1], [2], [3], [0], [1], [1], [2], [3]],
            [[3], [2], [1], [1], [0], [3], [2], [1], [1], [0]]
        ])
        np.testing.assert_array_equal(Y_out, expected)


from unittest.mock import MagicMock

import numpy as np
import numpy.testing as npt
import pandas as pd

from SleePyPhases.phases.BuildDataset import BuildDataset
from pyPhases.test.Mocks import OverwriteConfig
from pyPhases.test.TestCase import TestCase
from pyPhasesRecordloader import RecordLoader


class TestBuildDataset(TestCase):
    phase = BuildDataset()

    C1 = np.array([1, 2, 3, 4, 5])
    C2 = np.array([6, 7, 8, 9, 10])
    C3 = C2

    def config(self):
        self.project.setConfig("dataversion.split", {})
        return {
            "dataversion": {
                "seed": None,
                "split": {
                    "trainval": ["0:8"],
                    "test": ["8:10"],
                },
                "folds": 4,
                "recordIds": ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"],
                "groupBy": None,
            },
            "segmentManipulation": [{"name": "addBatchDimension"}, {"name": "fixedSize", "position": "center", "size": 10}],
            "segmentManipulationEval": [{"name": "addBatchDimension"}, {"name": "fixedSize", "position": "center", "size": 10}],
            "evalOn": None,
        }

    def setUp(self):
        super().setUp()

        self.project.registerData("metadata", pd.DataFrame([{"recordId": str(r)} for r in range(10)]))
        RecordLoader.getRecordList = MagicMock(return_value=["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"])
        
        self.project.registerData(
            "data-processed",
            np.array([np.arange(5).reshape(-1, 1) + i for i in range(10)]),
        )

        self.project.registerData(
            "data-features",
            np.array(
                [
                    np.array([0, 0, 0, 0, 0]).reshape(-1, 1),
                    np.array([0, 0, 0, 0, 1]).reshape(-1, 1),
                    np.array([0, 0, 0, 1, 0]).reshape(-1, 1),
                    np.array([0, 0, 0, 1, 1]).reshape(-1, 1),
                    np.array([0, 0, 1, 0, 0]).reshape(-1, 1),
                    np.array([0, 0, 1, 0, 1]).reshape(-1, 1),
                    np.array([0, 0, 1, 1, 0]).reshape(-1, 1),
                    np.array([0, 0, 1, 1, 1]).reshape(-1, 1),
                    np.array([0, 1, 0, 0, 0]).reshape(-1, 1),
                    np.array([0, 1, 0, 0, 1]).reshape(-1, 1),
                    # np.array([0, 1, 0, 1, 0]).reshape(-1, 1),
                    # np.array([0, 1, 0, 1, 1]).reshape(-1, 1),
                ]
            )
        )


    def testDataTest(self):
        dataset = self.getData("dataset-test")
        data = iter(dataset)

        self.assertEqual(len(dataset), 2)

        x, y = next(data)
        npt.assert_equal(x.reshape(-1), [0, 0, 0, 8, 9, 10, 11, 12, 0, 0])
        npt.assert_equal(y.reshape(-1), [-1, -1, -1, 0, 1, 0, 0, 0, -1, -1])

        x, y = next(data)
        npt.assert_equal(x.reshape(-1), [0, 0, 0, 9, 10, 11, 12, 13, 0, 0])
        npt.assert_equal(y.reshape(-1), [-1, -1, -1, 0, 1, 0, 0, 1, -1, -1])

    def testDataTraining(self):
        dataset = self.getData("dataset-training")

        self.assertEqual(len(dataset), 6)

        data = iter(dataset)
        x, y = next(data)
        npt.assert_equal(x.reshape(-1), [0, 0, 0, 2, 3, 4, 5, 6, 0, 0])
        npt.assert_equal(y.reshape(-1), [-1, -1, -1, 0, 0, 0, 1, 0, -1, -1])

        x, y = next(data)
        np.testing.assert_equal(x.reshape(-1), [0, 0, 0, 3, 4, 5, 6, 7, 0, 0])
        np.testing.assert_equal(y.reshape(-1), [-1, -1, -1, 0, 0, 0, 1, 1, -1, -1])

        x, y = next(data)
        np.testing.assert_equal(x.reshape(-1), [0, 0, 0, 4, 5, 6, 7, 8, 0, 0])
        np.testing.assert_equal(y.reshape(-1), [-1, -1, -1, 0, 0, 1, 0, 0, -1, -1])

        x, y = next(data)
        np.testing.assert_equal(x.reshape(-1), [0, 0, 0, 5, 6, 7, 8, 9, 0, 0])
        np.testing.assert_equal(y.reshape(-1), [-1, -1, -1, 0, 0, 1, 0, 1, -1, -1])

        x, y = next(data)
        np.testing.assert_equal(x.reshape(-1), [0, 0, 0, 6, 7, 8, 9, 10, 0, 0])
        np.testing.assert_equal(y.reshape(-1), [-1, -1, -1, 0, 0, 1, 1, 0, -1, -1])

        x, y = next(data)
        np.testing.assert_equal(x.reshape(-1), [0, 0, 0, 7, 8, 9, 10, 11, 0, 0])
        np.testing.assert_equal(y.reshape(-1), [-1, -1, -1, 0, 0, 1, 1, 1, -1, -1])

        # x, y = next(data)
        # np.testing.assert_equal(x.reshape(-1), [0, 0, 0, 8, 9, 10, 11, 12, 0, 0])
        # np.testing.assert_equal(y.reshape(-1), [-1, -1, -1, 0, 1, 0, 0, 0, -1, -1])

        # x, y = next(data)
        # np.testing.assert_equal(x.reshape(-1), [0, 0, 0, 9, 10, 11, 12, 13, 0, 0])
        # np.testing.assert_equal(y.reshape(-1), [-1, -1, -1, 0, 1, 0, 0, 1, -1, -1])

        self.assertRaises(StopIteration, data.__next__)

        data = iter(dataset)
        x, y = next(data)
        npt.assert_equal(x.reshape(-1), [0, 0, 0, 2, 3, 4, 5, 6, 0, 0])
        npt.assert_equal(y.reshape(-1), [-1, -1, -1, 0, 0, 0, 1, 0, -1, -1])

    @OverwriteConfig(fold=3)
    def testDataTrainingFold3(self):
        data = iter(self.getData("dataset-training"))

        x, y = next(data)
        npt.assert_equal(x.reshape(-1), [0, 0, 0, 0, 1, 2, 3, 4, 0, 0])
        npt.assert_equal(y.reshape(-1), [-1, -1, -1, 0, 0, 0, 0, 0, -1, -1])

        x, y = next(data)
        npt.assert_equal(x.reshape(-1), [0, 0, 0, 1, 2, 3, 4, 5, 0, 0])
        npt.assert_equal(y.reshape(-1), [-1, -1, -1, 0, 0, 0, 0, 1, -1, -1])

        x, y = next(data)
        npt.assert_equal(x.reshape(-1), [0, 0, 0, 2, 3, 4, 5, 6, 0, 0])
        npt.assert_equal(y.reshape(-1), [-1, -1, -1, 0, 0, 0, 1, 0, -1, -1])

        x, y = next(data)
        np.testing.assert_equal(x.reshape(-1), [0, 0, 0, 3, 4, 5, 6, 7, 0, 0])
        np.testing.assert_equal(y.reshape(-1), [-1, -1, -1, 0, 0, 0, 1, 1, -1, -1])

        x, y = next(data)
        np.testing.assert_equal(x.reshape(-1), [0, 0, 0, 4, 5, 6, 7, 8, 0, 0])
        np.testing.assert_equal(y.reshape(-1), [-1, -1, -1, 0, 0, 1, 0, 0, -1, -1])

        x, y = next(data)
        np.testing.assert_equal(x.reshape(-1), [0, 0, 0, 5, 6, 7, 8, 9, 0, 0])
        np.testing.assert_equal(y.reshape(-1), [-1, -1, -1, 0, 0, 1, 0, 1, -1, -1])

        self.assertRaises(StopIteration, data.__next__)

    def testDataValidation(self):
        data = iter(self.getData("dataset-validation"))

        x, y = next(data)
        npt.assert_equal(x.reshape(-1), [0, 0, 0, 0, 1, 2, 3, 4, 0, 0])
        npt.assert_equal(y.reshape(-1), [-1, -1, -1, 0, 0, 0, 0, 0, -1, -1])

        x, y = next(data)
        npt.assert_equal(x.reshape(-1), [0, 0, 0, 1, 2, 3, 4, 5, 0, 0])
        npt.assert_equal(y.reshape(-1), [-1, -1, -1, 0, 0, 0, 0, 1, -1, -1])

        self.assertRaises(StopIteration, data.__next__)

    @OverwriteConfig(fold=1)
    def testDataValidationFold1(self):
        data = iter(self.getData("dataset-validation"))

        x, y = next(data)
        npt.assert_equal(x.reshape(-1), [0, 0, 0, 2, 3, 4, 5, 6, 0, 0])
        npt.assert_equal(y.reshape(-1), [-1, -1, -1, 0, 0, 0, 1, 0, -1, -1])

        x, y = next(data)
        np.testing.assert_equal(x.reshape(-1), [0, 0, 0, 3, 4, 5, 6, 7, 0, 0])
        np.testing.assert_equal(y.reshape(-1), [-1, -1, -1, 0, 0, 0, 1, 1, -1, -1])

        self.assertRaises(StopIteration, data.__next__)

    @OverwriteConfig(fold=2)
    def testDataValidationFold2(self):
        data = iter(self.getData("dataset-validation"))

        x, y = next(data)
        np.testing.assert_equal(x.reshape(-1), [0, 0, 0, 4, 5, 6, 7, 8, 0, 0])
        np.testing.assert_equal(y.reshape(-1), [-1, -1, -1, 0, 0, 1, 0, 0, -1, -1])

        x, y = next(data)
        np.testing.assert_equal(x.reshape(-1), [0, 0, 0, 5, 6, 7, 8, 9, 0, 0])
        np.testing.assert_equal(y.reshape(-1), [-1, -1, -1, 0, 0, 1, 0, 1, -1, -1])

        self.assertRaises(StopIteration, data.__next__)

    @OverwriteConfig(fold=3)
    def testDataValidationFold3(self):
        data = iter(self.getData("dataset-validation"))

        x, y = next(data)
        np.testing.assert_equal(x.reshape(-1), [0, 0, 0, 6, 7, 8, 9, 10, 0, 0])
        np.testing.assert_equal(y.reshape(-1), [-1, -1, -1, 0, 0, 1, 1, 0, -1, -1])

        x, y = next(data)
        np.testing.assert_equal(x.reshape(-1), [0, 0, 0, 7, 8, 9, 10, 11, 0, 0])
        np.testing.assert_equal(y.reshape(-1), [-1, -1, -1, 0, 0, 1, 1, 1, -1, -1])

        self.assertRaises(StopIteration, data.__next__)



from typing import List
from pyPhases import Phase
from tqdm import tqdm
import pandas as pd

from pyPhasesRecordloader import AnnotationNotFound, RecordLoader, ChannelsNotPresent, Event, ParseError

from SleePyPhases.RecordFeatureExtraction import RecordFeatureExtraction
from SleePyPhases.SleepMetaData import SleepMetaData

class ExtractFeatures(Phase):
    allowTraining = False
    features = None

    def getChannelMetadata(self, recordId):
        recordLoader = RecordLoader.get()
        recordLoader.getSignalHeaders(recordId)

    def getRecordMetadata(self, recordId):
        recordLoader = RecordLoader.get()
        datasetConfig = self.getConfig("dataversion")
        metaData = {
            "recordId": recordId,
            "annotationExist": recordLoader.existAnnotation(recordId),
        }
        channelMetadata = []
        headers = recordLoader.getSignalHeaders(recordId)

        for header in headers:
            channelMetadata.append({
                "recordId": recordId,
                "channel": header["name"],
                "sample_rate": header["sample_rate"]
            })

        # if metaData["annotationExist"]:
        #     recordMetadata = recordLoader.getMetaData(recordId)
        #     metaData.update(recordMetadata)
        # else:
        #     self.logError(f"record id {recordId} does not exist")

        # if metaData["annotationExist"] and "minimalSamplingRate" in datasetConfig and datasetConfig["minimalSamplingRate"] is not None:
        #     minimalSamplingRates = datasetConfig["minimalSamplingRate"]

        #     try:
        #         headers = recordLoader.getSignalHeaders(recordId)

        #         for header in headers:
        #             typeStr = header["type"]
        #             if typeStr in minimalSamplingRates and header["sample_rate"] < minimalSamplingRates[typeStr]:
        #                 self.logError(f"record id {recordId} exluded because of minimal sampling rate for {typeStr}")
        #                 metaData["samplingRateCheck"] = False
        #                 break
        #     except ChannelsNotPresent as e:
        #         self.logError(f"record id {recordId} exluded because channels missinng {e.channels}")
        #         metaData["channelMissing"] = e.channels
        # return metaData
    
    def getFeature(self, name, recordId, options={}, channelType=None):
        recordLoader=RecordLoader.get()
        recordLoader.ignoreTargetSignals = True
        features = None
        recordDf = self.getData("metadata", pd.DataFrame)
        recordEntry = recordDf.loc[recordId]

        eventDf = self.getData("metadata-events", pd.DataFrame)

        featureExistName = f"feature-{name}"
        featureExist = featureExistName in recordDf.columns
        if not featureExist:
            recordDf[featureExistName] = False
        elif recordEntry[featureExistName]:
            return eventDf.query(f"recordId == '{recordId}' and feature == '{name}'")

        # calcluate Features
        try:
            recordSignal = recordLoader.getHarmonizedSignal(recordId)

            events: List[Event] = RecordFeatureExtraction.getInstance().step(name, recordSignal, *options)

            if channelType is not None:
                channelType = [s.name for s in recordSignal.signals if s.typeStr == channelType]
            else:
                channelType = [None]
                events = [events]

            eventList = []
            for index, ev in enumerate(events):
                eventList += [
                    {
                        "recordId": recordId,
                        "channel": channelType[index],
                        "feature": name,
                        "event": e.name,
                        "start": e.start,
                        "duration": e.duration
                    } for e in ev
                ]
            featureDf = pd.DataFrame(eventList)
            eventDf = pd.concat((eventDf, featureDf))
            self.registerData("metadata-events", eventDf)
            recordDf.loc[recordId, featureExistName] = True
            self.registerData("metadata", recordDf)

            return eventDf

            # features = FeatureExtraction(self.project).step(name, recordSignal)
        except ChannelsNotPresent as e:
            self.logError(f"channels not present: {e.channels}")
            pass

        return None
        

    def getSomnometrics(self, name, recordId):
        metricConfig = self.getConfig(f"somnometrics.{name}", {})
        metricType = metricConfig["type"] if "type" in metricConfig else "recordFeature"

        metadataRecords = self.getData("metadata", pd.DataFrame)
        metaDataChannels = self.getData("metadata-channels", pd.DataFrame)

        # metadataRecords = pd.DataFrame.from_records(metadataRecords, index="recordId")
        # metaDataChannels = pd.DataFrame.from_records(metaDataChannels)
 
        if recordId not in metadataRecords.index:
            recordMetadata = self.getRecordMetadata(recordId)
            metadataRecords = metadataRecords.append(recordMetadata)
            self.registerData("metadata", metadataRecords)
        else:
            recordMetadata = metadataRecords.query(f"recordId == '{recordId}'")

        lookupDf = metaDataChannels if metricType == "recordChannel" else metadataRecords
        metaDataEntries = lookupDf.query(f"recordId == '{recordId}'")
            
        sm = SleepMetaData()
        if name not in metaDataEntries.columns:
            metricConfig = self.getConfig("somnometrics")[name]
            options = metricConfig["options"] if "options" in metricConfig else {}
            channelType = metricConfig["channelType"] if "channelType" in metricConfig else {}
            
            featureNames = metricConfig["features"]

            for featureName in featureNames:
                featureList = self.getFeature(featureName, recordId, options, channelType=channelType)

                sm.addFeatureList(featureName, featureList)

        return sm.getMetaData(name)

    # featureConfigs is a dict with feature names as keys, so that if they exist this method is not called and loaded from the feature table directly
    # if there is no featureConfig all features will be calculated for a record, to optimize the record loading
    # feature name is used to recalculate the feature, so it will always return the full recording row, but only recalculate if the feature is not present
    def generateData(self, dataId, recordId=None, channel=None, recalculate=True, **featureNameDict):
        if dataId != "features":
            raise Exception(f"data id {dataId} not supported")

        if recordId is None:
            # self.features = pd.DataFrame(columns=["recordId", "channel"]).set_index(["recordId", "channel"], drop=False)
            self.features = pd.DataFrame(index=pd.MultiIndex(levels=[[],[]], codes=[[],[]], names=['recordId','channel']))
            self.registerData("features", self.features, save=False)
            return self.features

        if self.features is None:
            self.features = self.getData("features", pd.DataFrame)

        featureConfigs = self.getConfig("featureExtraction.features", [])
        if featureNameDict != {}:
            featureNames = list(featureNameDict.keys())
            featureConfigs = [f for f in featureConfigs if f["name"] in featureNames]

        recordLoader = RecordLoader.get()
        recordSignal = None
        try:
            recordSignal, eventlist = recordLoader.loadRecord(recordId)
        except AnnotationNotFound as e:
            self.logError(f"Annotation missing in {recordId}: {e.name} ... skipping record")
        except ChannelsNotPresent as e:
            self.logError(f"Channels missing in {recordId}: {e.channels} ... skipping record")
        except ParseError as e:
            self.logError(f"ParseError in record {recordId}, record will be skipped: {e}")

            
        for featureConfig in featureConfigs:
            if recordSignal is None:
                self.features.loc[recordId, "recordId"] = recordId
                self.features.loc[recordId, featureConfig["name"]] = True
                self.registerData("features", self.features)
                return {}
            
            channels = featureConfig["channels"] if channel is None else [channel]
            for c in channels:
                try:
                    idx = (recordId, c)
                    featureStep = featureConfig["type"]
                    RecordFeatureExtraction.getInstance().updateMetadataByStep(featureStep, recordSignal, eventlist, metadata=self.features, channel=c)
                except ChannelsNotPresent as e:
                    self.logError(f"channels not present: {e.channels}")

                self.features.loc[idx, "recordId"] = recordId
                self.features.loc[idx, "channel"] = c
                self.features.loc[idx, featureConfig["name"]] = True
        
        self.registerData("features", self.features)
        idx = (recordId, channel) if channel is not None else recordId
        return self.features.loc[idx]

    def main(self):
        import pandas as pd
        self.features = None
        metadata = self.project.getData("metadata", pd.DataFrame)
        recordIds = metadata["recordId"]
        forceCalculation = self.getConfig("featureExtraction.force", False)

        reqFeatures = {f["name"]: True for f in self.getConfig("featureExtraction.features", [])}

        recordId = self.getConfig("recordId", False)
        recordIds = [recordId] if recordId else recordIds.iloc

        for recordId in tqdm(recordIds):
            if forceCalculation:
                # the parameter forced has no meaning, it is just known to be not in the metadata, so the record will always be recalculated
                reqFeatures["forced"] = True # should never be found
            rows = self.getData("features", pd.DataFrame, recordId=recordId, **reqFeatures)
            if len(rows) == 0:
                self.logError(f"no features for {recordId}")
                



from pathlib import Path
from pyPhases import Phase

from pyPhasesML import ModelManager
from pyPhasesRecordloader import RecordLoader, ChannelsNotPresent

class VisualizeConfig(Phase):

    def main(self):
        config = self.project.config
        verbosity = 1

        self.log("Pipeline Summary")
        self.log("===========================================")

        recordLoader = RecordLoader.get()
        # Data loading
        self.log("\n############## Data Loading ##############")
        self.log(f"Database: {config['dataBase']}")
        # self.log(f"   Database Version: {config['dataBaseVersion']}")

        self.log(f"Loader: {config['useLoader']}")
        recordLoaderClass = type(recordLoader).__name__
        self.log(f"RecordLoader: {recordLoaderClass}")
        self.log(f"Dataset Path: {config['filePath']} - exists: {Path(config['filePath']).exists()}")

        # Record filtering
        # self.log("")
        self.log("\n############## Record Filtering ##############")
        dataversion = config["dataversion"]

        if "recordIds" in dataversion and dataversion["recordIds"] is not None:
            self.log(f"Specific Record IDs: {dataversion['recordIds']}")
            self.log(f"Record Count: {len(dataversion['recordIds'])}")
        else:
           self.log(f"Datasplit:")
           self.log(dataversion)
           self.log(f"{dataversion['split']}")
           recordIds = self.getData("allDBRecordIds", generate=False)
           self.log(f"   RecordCount: {len(recordIds)}")

        # # Data preprocessing
        # self.log("\n3. Data Preprocessing:")
        # self.log(f"   Input Shape: {config['inputShape']}")
        # self.log("   Segment Manipulations:")
        # for manipulation in config['segmentManipulation']:
        #     self.log(f"   - {manipulation['name']}")

        # Channel preprocessing
        # self.log("")
        self.log("\n############## Channel Preprocessing ################")
        
        if verbosity < 2:
            self.log(f"Available channels in the dataset: {recordLoader.targetSignals}")

        self.log("Channels:")
        preprocessingConfig = config['preprocessing']
        usedTypes = []
        self.log(f"Target channel samplingrate: {preprocessingConfig['targetFrequency']}")
        
        for index, channels in enumerate(preprocessingConfig['targetChannels']):
            channelFound = False
            for channelName in channels:
                try:
                    channelType = recordLoader.getSignalTypeByName(channelName)
                    self.log(f"Channel {index+1}: (one of {channels}): {channelName} - Type: {channelType}")
                    channelFound = True
                    usedTypes.append(channelType)
                    break # only the first channel will be processed
                except ChannelsNotPresent:
                    continue

            if not channelFound:
                self.logError(f"Channel {index+1}: No channels found for {channels} Available Channels in the dataset: {recordLoader.targetSignals}")

                
            # self.log("")
            # self.log(f"   - {channel['name']} (Type: {channel['type']})")
        # self.log("")
        self.log("\n############## Labels ##############")
        self.log(f"Target label samplingrate: {preprocessingConfig['labelFrequency']}")
        self.log(f"Label Channels: {config['labelChannels']}")
        self.log(f"Event Manipulation: {preprocessingConfig['extendEvents']}")


        # self.log("")
        self.log("\n############## Preprocessing Steps ##############")
        for channelType in usedTypes:
            steps = config['preprocessing']['stepsPerType'][channelType]
            stepsStr = ""
            if isinstance(steps, list):
                stepsStr = [step['name'] for step in steps]
            else: # is dict
                stepsStr = [step['name'] for step in steps.values()]
            self.log(f"Channel-Type {channelType}: {stepsStr}")

        
        # self.log("")
        self.log("\n############## Data Manipulation ##############")
        inputType = "record wise" if config["recordWise"] else "segment wise"
        self.log(f"Manipulation Input: {inputType}")
        if not config["recordWise"]:
            self.log(f"   Segment Size Channel: {config['segmentLength']}")
            self.log(f"   Segment Size Label: {config['segmentLengthLabel']}")
        
        stepsTraining = [s["name"] for s in config["segmentManipulation"]]
        stepsValidation = [s["name"] for s in config["segmentManipulationEval"]]

        stepsBatch = [s["name"] for s in config["batchManipulation"]]

        self.log(f"Steps (Training): {stepsTraining}")
        self.log(f"Steps (Validation/Test): {stepsValidation}")
        self.log(f"Steps (Batch): {stepsBatch}")
        # Model
        # self.log("")
        self.log("\n############## Model ##############")
        self.log(f"Model Name: {config['modelName']}")
        self.log(f"Input Shape: {config['inputShape']}")

        # self.log("")
        self.log("\n############## Training ##############")
        trainingParameter = self.getConfig("trainingParameter")

        model = ModelManager().getModel()
        lossFunction = model.getLossFunction()
        lossFunction = lossFunction if isinstance(lossFunction, str) else type(lossFunction).__name__

        self.log(f"Batch Size: {trainingParameter['batchSize']}")
        self.log(f"Learningrate: {trainingParameter['learningRate']}")
        self.log(f"Learningrate Decay: {trainingParameter['learningRateDecay']}")
        self.log(f"Cyclic Learningrate: {trainingParameter['cyclicLearningRate']}")

        self.log(f"Loss Function: {lossFunction}")
        self.log(f"Optimizer: {trainingParameter['optimizer']}")
        self.log(f"Early Stopping: {trainingParameter['stopAfterNotImproving']}")
        self.log(f"Max Epochs: {trainingParameter['maxEpochs']}")
        # model = ModelManager().getModel()
        # torchModel = model.model
        # outputShape = torchModel.output_shape

        self.log("\n############## Output ##############")
        self.log(f"Number of Labels: {config['numLabels']}")
        self.log(f"Labels: {config['classification']['labelNames']}")
        self.log(f"Classes: {config['classification']['classNames']}")

        self.log("\n############## Postprocessing ##############")
        postSteps =  [s["name"] for s in config["manipulationAfterPredict"]]
        self.log(f"Postprocessing Steps: {postSteps}")
        thresholdsStr = f"fixed: {config['fixedThresholds']}" if "fixedThresholds" in config else f"will be optimized, using {config['thresholdMetric']} on {config['optimizeOn']}" 
        self.log(f"Thresholds: {thresholdsStr}")
        # self.log("   Model Architecture:")
        # pself.log.pself.log(config['model'], indent=4)

        # # Output
        # self.log("\n6. Output:")
        # self.log(f"   Number of Classes: {config['numClasses']}")
        # self.log(f"   Classification: {config['classification']['name']}")
        # self.log("   Class Names:")
        # for class_name in config['classification']['classNames']:
        #     self.log(f"   - {class_name}")


import contextlib
from pathlib import Path

import numpy as np
from pyPhases import Phase
from pyPhasesML import Model, ModelManager, TrainingSetLoader
from pyPhasesML.adapter.torch import CyclicLearningrate, FindLearningRate


class Training(Phase):
    def prepareConfig(self):
        shape = self.getConfig("inputShape")  # [segmentLength, channelCount]

        self.log(f"Set the inputshape to {str(shape)}")
        self.setConfig("inputShape", shape)

    def createLogFolder(self, path):
        if not Path(path).exists():
            Path(path).mkdir(parents=True, exist_ok=True)
        # you can throw an error if you want to prevent overwriting you training

    def train(self):
        model = ModelManager.getModel()
        # log everything in a specific folder derived from config values
        modelConfigString = self.project.getDataFromName("modelState").getTagString()
        logPath = f"logs/{modelConfigString}/"
        self.createLogFolder(logPath)
        model.config.logPath = logPath

        # load the trainingsdata from filesystem or generate it
        train = self.project.generateData("dataset-training")
        val = self.project.generateData("dataset-validation")
        trainingsSet = TrainingSetLoader(trainingData=train, validationData=val)

        findLR = self.getConfig("trainingParameter.findCyclicLearningRate", False)
        cyclicLR = self.getConfig("trainingParameter.cyclicLearningRate", False)

        if findLR:
            findinLR = FindLearningRate(model.config, minLR=0.00001, maxLR=2, iterations=3)
            model.registerCB(findinLR)
        elif cyclicLR:
            model.registerCB(CyclicLearningrate(model.config))

        self.log(model.summary())
        self.project.saveConfig(f"{logPath}model.config", "modelState")
        self.project.saveConfig(f"{logPath}project.config")
        trainedModel = model.train(trainingsSet)

        # save the model state and relevant config values
        self.logSuccess(f"Model trained and saved to {logPath}")
        self.logSuccess(f"Model trained {model.validationMetrics[0]}: {model.bestMetric}")

        self.gridOutput = {
            "epochs": model.fullEpochs,
            "best": model.bestMetric,
        }
        return trainedModel

    def generateData(self, name):
        trainedModel = self.train()
        self.project.registerData("modelState", trainedModel)
        self.project.registerData("modelStateConfig", self.project.getDataFromName("modelState").getDependencyDict())

    def main(self):
        startfold = self.getConfig("startFold", 0)
        folds = max(self.getConfig("dataversion.folds", 1), 1)
        endfold = self.getConfig("endFold", folds)
        self.gridOutput = {
            "epochs": 0,
            "best": 0,
        }

        if startfold >= folds:
            self.logError(f"startFold {startfold} is bigger than the number of folds {folds}")
        if endfold <= startfold:
            self.logError(f"endfold {endfold} is smaller than start folds {startfold}")

        gridOutputs = []
        for fold in range(startfold, endfold):
            # save model state (random initialisation or pretrained)
            startState = ModelManager.getModel().model.state_dict()
            self.setConfig("fold", fold)
            # only train if it doesnt allready exist
            self.getData("modelState", Model)
            # unregister current model for the next fold
            with contextlib.suppress(KeyError):
                self.project.unregister("modelState")
                self.project.unregister("modelStateConfig")
            gridOutputs.append(self.gridOutput)
            # reset model state
            ModelManager.getModel().model.load_state_dict(startState)

        self.project.gridOutput = {
            "epochs": str(np.mean([m["epochs"] for m in gridOutputs])),
            "best": str(np.mean([m["best"] for m in gridOutputs])),
            "folds": len(gridOutputs),
        }



from pyPhases import Phase
from pyPhasesML import Model, ModelManager, Scorer
from tqdm import tqdm

class Validation(Phase):
    allowTraining = False

    def getModel(self) -> Model:
        modelState = self.project.getData("modelState", Model, generate=self.allowTraining)
        model = ModelManager.getModel(True)
        model.build()
        model.loadState(modelState)
        return model

    def main(self):
        scorer = Scorer(classNames=self.getConfig("classification.labelNames"), trace=True)
        scorer.metrics = self.getConfig("trainingParameter.validationMetrics")
        scorer.ignoreClasses = [self.getConfig("ignoreClassIndex")]

        model = self.getModel()
        validationData = self.project.generateData("dataset-validation")

        for x, t in tqdm(validationData):
            p = model.predict(x, get_likelihood=True, returnNumpy=True)
            scorer.score(t, p, trace=True)

        res = scorer.scoreAllRecords()
        self.logSuccess(f"Validation results: {res}")
        
        self.project.gridOutput = res


# import contextlib
# from unittest.mock import MagicMock
# import numpy as np
# import numpy.testing as npt
# from pyPhases.test.Mocks import OverwriteConfig
# from pyPhases.test.TestCase import TestCase
# from pyPhasesRecordloader import RecordSignal, Signal, RecordLoader, SignalPreprocessing

# from SleePyPhases.phases.Eval import Eval



# class TestEval(TestCase):
#     phase = Eval()

#     C1 = np.array([1, 2, 3, 4, 5])
#     C2 = np.array([6, 7, 8, 9, 10])
#     C3 = C2

#     # def config(self):
#     #     return   eventEval
#     # :
#     # manipulationAfterPredict: # postprocessing
#     # - name: deleteIgnoredMultiClass
#     # - name: reduceY
#     #   factor: 750
#     #   reduce: max}

#     def setUp(self):
#         super().setUp()

#         class MyPreprocessing(SignalPreprocessing):
#             def add1(signal: Signal, recordSignal: RecordSignal, config: dict):
#                 signal.signal += 1

#         self.preprocessing = MyPreprocessing(self.getConfig("preprocessing"))

#         dataProcessed = np.array(
#             [
#                 self.C1.reshape(-1, 1),
#                 self.C2.reshape(-1, 1),
#             ]
#         )

#         dataFeatures = np.array(
#             [
#                 np.array([0, 0, 1, 1, 0]).reshape(-1, 1),
#                 np.array([0, 1, 0, 0, 0]).reshape(-1, 1),
#             ]
#         )
#         self.project.registerData("metadata", [{"recordId": str(r)} for r in range(10)])
#         RecordLoader.getRecordList = MagicMock(return_value=["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"])

#         self.project.setConfig("datasetSplit", "test")
#         self.project.registerData("data-processed", dataProcessed)
#         self.project.registerData("data-features", dataFeatures)

#         self.project.setConfig("datasetSplit", "training")
#         self.project.registerData("data-processed", dataProcessed + 1)
#         self.project.registerData("data-features", dataFeatures + 1)

#         self.project.setConfig("datasetSplit", "validation")
#         self.project.registerData("data-processed", dataProcessed + 2)
#         self.project.registerData("data-features", dataFeatures + 2)

#     @OverwriteConfig({"datasetSplit": "test"})
#     def testDataTest(self):
#         data = iter(self.getData("dataset-test"))

#         x, y = data.__next__()
#         npt.assert_equal(x.reshape(-1), [0, 0, 0, 1, 2, 3, 4, 5, 0, 0])
#         npt.assert_equal(y.reshape(-1), [-1, -1, -1, 0, 0, 1, 1, 0, -1, -1])

#         x, y = data.__next__()
#         npt.assert_equal(x.reshape(-1), [0, 0, 0, 6, 7, 8, 9, 10, 0, 0])
#         npt.assert_equal(y.reshape(-1), [-1, -1, -1, 0, 1, 0, 0, 0, -1, -1])

#     @OverwriteConfig({"datasetSplit": "training"})
#     def testDataTraining(self):
#         data = iter(self.getData("dataset-training"))

#         x, y = data.__next__()
#         npt.assert_equal(x.reshape(-1), [0, 0, 0, 2, 3, 4, 5, 6, 0, 0])
#         npt.assert_equal(y.reshape(-1), [-1, -1, -1, 1, 1, 2, 2, 1, -1, -1])

#         x, y = data.__next__()
#         npt.assert_equal(x.reshape(-1), [0, 0, 0, 7, 8, 9, 10, 11, 0, 0])
#         npt.assert_equal(y.reshape(-1), [-1, -1, -1, 1, 2, 1, 1, 1, -1, -1])

#     @OverwriteConfig({"datasetSplit": "validation"})
#     def testDataValidation(self):
#         data = iter(self.getData("dataset-validation"))

#         x, y = data.__next__()
#         npt.assert_equal(x.reshape(-1), [0, 0, 0, 3, 4, 5, 6, 7, 0, 0])
#         npt.assert_equal(y.reshape(-1), [-1, -1, -1, 2, 2, 3, 3, 2, -1, -1])

#         x, y = data.__next__()
#         npt.assert_equal(x.reshape(-1), [0, 0, 0, 8, 9, 10, 11, 12, 0, 0])
#         npt.assert_equal(y.reshape(-1), [-1, -1, -1, 2, 3, 2, 2, 2, -1, -1])

#     @OverwriteConfig({"BuildDataset": {"useMultiThreading": True, "threads": 2}, "datasetSplit": "validation"})
#     def testMultiThreading(self):
#         with contextlib.suppress(KeyError):
#             self.project.unregister("dataset-validation")
#         data = self.getData("dataset-validation")
#         datagen = iter(data)
#         x, y = next(datagen)
#         npt.assert_equal(x.reshape(-1), [0, 0, 0, 3, 4, 5, 6, 7, 0, 0])
#         npt.assert_equal(y.reshape(-1), [-1, -1, -1, 2, 2, 3, 3, 2, -1, -1])

#         x, y = next(datagen)
#         npt.assert_equal(x.reshape(-1), [0, 0, 0, 8, 9, 10, 11, 12, 0, 0])
#         npt.assert_equal(y.reshape(-1), [-1, -1, -1, 2, 3, 2, 2, 2, -1, -1])
